# -*- coding: utf-8 -*-
'film_distance_relaxation': False,   # if True, sets relaxXYZ="FFT" for all atoms
'force_criterion': 0.049,            # Sets the threshold of the largest force
'relax_iter': 5                      # Maximum number of optimization iterations



###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/JuDFTteam/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################
"""
In this file/module, YOU, the user can specify some default resource values for
queues of different computers
This might be really useful for high-throughput calculation.
You can modify, adjust this file to your needs
"""

# TODO: move computers dict somewhere else?
# TODO find AiiDA solution for this


def queue_defaults(queue_name, computer=None):
    """
    In this class you specify defaults methods which you can use for workflows
    or normal calculations.
    """
    '''
    code = Code.get_from_string(codename)
    code2 = Code.get_from_string(codename2)
    computer = Computer.get(computer_n
    ame)
    '''
    queue_resources = None
    print(queue_name)
    computers = {
        'iff003': {
            'th1': {
                'resources': {
                    'num_machines': 1,
                    'num_mpiprocs_per_machine': 12
                },
                'walltime_sec': 30 * 60
            },
            'th1_small': {
                'resources': {
                    'num_machines': 1,
                    'num_mpiprocs_per_machine': 12
                },
                'walltime_sec': 20 * 60
            }
        }
    }

    if computer:
        #print 'computer'
        c_name = computer
        queue = computers.get(c_name, {}).get(queue_name, {})
        res = queue.get('resources', None)
        wtime = queue.get('walltime_sec', None)
    else:
        #print 'no computer'
        c_name = None
        res = None
        wtime = None
        for comp, queues in computers.items():
            queue = queues.get(queue_name, {})
            # queues might not be unique to certain computers overall
            # but this is the users responsibility
            # print 'queue {}'.format(queue)
            if queue:
                res = queue.get('resources', None)
                wtime = queue.get('walltime_sec', None)

    queue_resources = {'resources': res, 'walltime_sec': wtime}
    #print queue_resources

    return queue_resources



###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/JuDFTteam/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################
"""
This module, contains a method to merge Dict nodes used by the FLEUR inpgen.
This might also be of interest for other all-electron codes
"""
# Shall we allow for a python dictionary also instead of forcing paramteraData?
# but then we can not keep the provenace...

from aiida.plugins import DataFactory
from aiida.orm import Bool, Dict
from aiida.engine import calcfunction as cf
#Dict = DataFactory('dict')


def merge_parameter(Dict1, Dict2, overwrite=True, merge=True):
    """
    Merges two Dict nodes.
    Additive: uses all namelists of both.
    If they have a namelist in common. Dict2 will overwrite the namelist
    of Dict. If this is not wanted. set overwrite = False.
    Then attributes of both will be added, but attributes from Dict1 won't
    be overwritten.


    :param Dict1: AiiDA Dict Node
    :param Dict2: AiiDA Dict Node
    :param overwrite: bool, default True
    :param merge: bool, default True

    returns: AiiDA Dict Node

    #TODO be more carefull how to merge ids in atom namelists, i.e species labels
    """

    from aiida.common.exceptions import InputValidationError
    from aiida_fleur.tools.dict_util import recursive_merge
    #Dict = DataFactory('dict')

    # layout:
    # check input
    # get dictionaries
    # merge dictionaries into new dictionary
    # create a new Dict node

    new_dict = {}
    atoms_dict = {}
    atomlist = []
    if not isinstance(Dict1, Dict):
        raise InputValidationError('Dict1, must be of type Dict')
    if not isinstance(Dict2, Dict):
        raise InputValidationError('Dict2, must be of type Dict')
    dict1 = Dict1.get_dict()
    dict2 = Dict2.get_dict()

    if dict1 == dict2:
        return Dict(dict=dict1)

    for key in list(dict1.keys()):
        if 'atom' in key:
            val = dict1.pop(key)
            atomlist.append(val)

    for key in list(dict2.keys()):
        if 'atom' in key:
            val = dict2.pop(key)
            atomlist.append(val)

    # TODO do something on atom list,
    # we do not want doubles, check element and Id? Keep first ones?

    for i, atom in enumerate(atomlist):
        # TODO check for duplicates? what about
        key = f'atom{i}'
        atoms_dict[key] = atom

    # merge all namelists except atoms
    if overwrite:
        new_dict = dict1.copy()
        new_dict.update(dict2)
    else:
        # add second one later?
        new_dict = dict2.copy()
        if merge:
            new_dict = recursive_merge(new_dict, dict1)
        else:
            new_dict.update(dict1)
        # TODO mergeing does not make sense for all namelist keys.
        # be more specific here.
    new_dict.update(atoms_dict)

    # be carefull with atom namelist

    return Dict(dict=new_dict)


def merge_parameters(DictList, overwrite=True):
    """
    Merge together all parameter nodes in the given list.
    """
    #Dict = DataFactory('dict')
    paremeter_data_new = Dict(dict={})

    for i, parameter in enumerate(DictList):
        if isinstance(parameter, Dict):
            # merge
            paremeter_data_new = merge_parameter(paremeter_data_new, parameter, overwrite=overwrite)
        else:
            print(f'WARNING: Entry : {i} {parameter} is not of type Dict, I skip it.')

    return paremeter_data_new


@cf
def merge_parameter_cf(Dict1, Dict2, overwrite=None):
    """
    calcfunction of merge_parameters
    """
    if overwrite is None:
        overwrite = Bool(True)
    paremeter_data_new = merge_parameter(Dict1, Dict2, overwrite=overwrite)

    return paremeter_data_new


'''
# TODO how to deal with a list? *args, prob is not the best, also it is not working here.
# makeing a methods m(self, *args, **kwargs) and setting some fallbacks, does not work, because self, cannot be parsed
# I guess...
@cf
def merge_parameters_wf(*Dicts, overwrite=Bool(True)):
    """
    calcfunction of merge_parameters
    """
    DictList = []
    for parameter in Dicts:
        DictList.append(parameter)
    paremeter_data_new = merge_parameters(DictList, overwrite=overwrite)

    return paremeter_data_new
'''
'''
#TODO this has to moved into cmdline
if __name__ == '__main__':
    import argparse
    #Dict = DataFactory('dict')

    parser = argparse.ArgumentParser(description='Merge a Dict node.')
    parser.add_argument('--para1', type=Dict, dest='para1', help='The first Dict node', required=True)
    parser.add_argument('--para2', type=Dict, dest='para2', help='The second Dict node', required=True)
    parser.add_argument('--overwrite',
                        type=bool,
                        dest='overwrite',
                        help='Shall values given in Dict2 overwrite the values from the first Dict?',
                        required=False)
    args = parser.parse_args()
    merge_parameter(Dict1=args.para1, Dict2=args.para1, overwrite=args.overwrite)
    '''



###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/JuDFTteam/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################
"""
    In this module you find the workflow 'FleurCreateMagneticWorkChain' for creation of relaxed
    film deposited on a cubic substrate.
"""

import copy

from aiida.engine import WorkChain, if_
from aiida.engine import calcfunction as cf
from aiida.orm import StructureData, Dict, Float, load_node, Str
from aiida.common import AttributeDict
from aiida.common.exceptions import NotExistent

from aiida_fleur.workflows.eos import FleurEosWorkChain
from aiida_fleur.workflows.base_relax import FleurBaseRelaxWorkChain


class FleurCreateMagneticWorkChain(WorkChain):
    """
        This workflow creates relaxed magnetic film on a substrate.
    """
    _workflowversion = '0.2.1'

    _default_wf_para = {
        'lattice': 'fcc',
        'miller': None,
        'directions': None,
        'host_symbol': 'Pt',
        'latticeconstant': 4.0,  # if equals to 0, use distance_suggestion
        'size': (1, 1, 5),
        'replacements': {
            0: 'Fe',
            -1: 'Fe'
        },
        'hold_layers': None,
        'last_layer_factor': 0.85,
        'first_layer_factor': 0.0,
        'adjustment_needed': True,
        'decimals': 10,
        'z_coordinate_window': 2,
        'pop_last_layers': 1,
        'AFM_name': 'FM',
        'magnetic_layers': 1,
        'AFM_layer_positions': None,
        'total_number_layers': 4,
        'num_relaxed_layers': 2
    }

    @classmethod
    def define(cls, spec):
        super().define(spec)
        spec.expose_inputs(FleurEosWorkChain,
                           namespace_options={
                               'required': False,
                               'populate_defaults': False
                           },
                           namespace='eos',
                           exclude=('structure',))
        spec.expose_inputs(FleurBaseRelaxWorkChain,
                           namespace_options={
                               'required': False,
                               'populate_defaults': False
                           },
                           namespace='relax',
                           exclude=('structure',))
        spec.input('wf_parameters', valid_type=Dict, required=False)
        spec.input('eos_output', valid_type=Dict, required=False)
        spec.input('optimized_structure', valid_type=StructureData, required=False)
        spec.input('distance_suggestion', valid_type=Dict, required=False)

        spec.outline(cls.start,
                     if_(cls.eos_needed)(cls.run_eos,),
                     if_(cls.relax_needed)(cls.run_relax,), cls.make_magnetic)

        spec.output('magnetic_structure', valid_type=StructureData)

        # exit codes
        spec.exit_code(230, 'ERROR_INVALID_INPUT_PARAM', message='Invalid workchain parameters.')
        spec.exit_code(231, 'ERROR_INVALID_INPUT_CONFIG', message='Invalid input configuration.')
        spec.exit_code(380, 'ERROR_NOT_SUPPORTED_LATTICE', message='Specified substrate has to be bcc or fcc.')
        spec.exit_code(382, 'ERROR_RELAX_FAILED', message='Relaxation calculation failed.')
        spec.exit_code(383, 'ERROR_EOS_FAILED', message='EOS WorkChain failed.')

    def eos_needed(self):
        """
        Returns True if EOS WorkChain should be submitted
        """
        return self.ctx.eos_needed

    def prepare_eos(self):
        """
        Initialize inputs for eos workflow:
        wf_param, options, calculation parameters, codes, structure
        """
        inputs = AttributeDict(self.exposed_inputs(FleurEosWorkChain, namespace='eos'))
        inputs.metadata.label = 'EOS_substrate'
        inputs.metadata.description = 'The EOS workchain finding equilibrium substrate'
        # Here wf_dict nodes appears out of nowwhere.
        inputs.structure = create_substrate_bulk(Dict(dict=self.ctx.wf_dict))

        if not isinstance(inputs.structure, StructureData):
            return inputs, inputs.structure  # exit code thrown in create_substrate_bulk

        return inputs, None

    def run_eos(self):
        """
        Optimize lattice parameter for substrate bulk structure.
        """
        self.report('INFO: submit EOS WorkChain')
        inputs = {}
        inputs, error = self.prepare_eos()
        if error:
            return error
        res = self.submit(FleurEosWorkChain, **inputs)
        self.to_context(eos_wc=res)

    def start(self):
        """
        Retrieve and initialize paramters of the WorkChain
        """
        self.report(f'INFO: started Create Magnetic Film workflow version {self._workflowversion}\n')

        self.ctx.info = []
        self.ctx.warnings = []
        self.ctx.errors = []
        self.ctx.energy_dict = []
        self.ctx.substrate = None

        # initialize the dictionary using defaults if no wf paramters are given
        wf_default = copy.deepcopy(self._default_wf_para)
        if 'wf_parameters' in self.inputs:
            wf_dict = self.inputs.wf_parameters.get_dict()
        else:
            wf_dict = wf_default

        extra_keys = []
        for key in wf_dict.keys():
            if key not in list(wf_default.keys()):
                extra_keys.append(key)
        if extra_keys:
            error = f'ERROR: input wf_parameters for Create Magnetic contains extra keys: {extra_keys}'
            self.report(error)
            return self.exit_codes.ERROR_INVALID_INPUT_PARAM

        # extend wf parameters given by user using defaults
        for key, val in wf_default.items():
            wf_dict[key] = wf_dict.get(key, val)
        self.ctx.wf_dict = wf_dict

        inputs = self.inputs
        if 'eos' in inputs:
            self.report('INFO: EOS workchain will be submitted')
            self.ctx.eos_needed = True
            self.ctx.relax_needed = True
            if 'eos_output' in inputs:
                self.report('ERROR: you specified both eos_output and eos wc inputs.')
                return self.exit_codes.ERROR_INVALID_INPUT_CONFIG
            if 'relax' not in inputs:
                self.report('ERROR: no relax wc input was given despite EOS is needed.')
                return self.exit_codes.ERROR_INVALID_INPUT_CONFIG
            if 'optimized_structure' in inputs:
                self.report('ERROR: optimized structure was given despite EOS is needed.')
                return self.exit_codes.ERROR_INVALID_INPUT_CONFIG
        else:
            if 'eos_output' in inputs:
                self.report('INFO: Outputs of the given EOS workchain will be used for relaxation')
                self.ctx.eos_needed = False
                self.ctx.relax_needed = True
                if 'relax' not in inputs:
                    self.report('ERROR: no relax wc input was given despite EOS is needed.')
                    return self.exit_codes.ERROR_INVALID_INPUT_CONFIG
                if 'optimized_structure' in inputs:
                    self.report('ERROR: optimized structure was given despite relax is needed.')
                    return self.exit_codes.ERROR_INVALID_INPUT_CONFIG
            else:
                if 'optimized_structure' in inputs:
                    self.report('INFO: given relaxed structure will be used, no EOS or relax WC')
                    self.ctx.eos_needed = False
                    self.ctx.relax_needed = False
                    if 'relax' in inputs:
                        self.report('ERROR: relax wc input was given but relax is not needed.')
                        return self.exit_codes.ERROR_INVALID_INPUT_CONFIG
                else:
                    self.ctx.eos_needed = False
                    self.ctx.relax_needed = True
                    self.report('INFO: relaxation will be continued; no EOS')
                    if 'relax' not in inputs:
                        self.report('ERROR: relax wc input was not given but relax is needed.')
                        return self.exit_codes.ERROR_INVALID_INPUT_CONFIG

        if 'relax' in inputs and 'distance_suggestion' not in inputs:
            if 'eos' in inputs or 'eos_output' in inputs:
                self.report('ERROR: relax wc input was given but distance_suggestion was not.')
                return self.exit_codes.ERROR_INVALID_INPUT_CONFIG

        if self.ctx.wf_dict['latticeconstant'] == 0 and 'distance_suggestion' not in inputs:
            self.report('ERROR: latticeconstant equals to 0 but distance_suggestion was not given.')
            return self.exit_codes.ERROR_INVALID_INPUT_CONFIG

        if not self.ctx.wf_dict['latticeconstant']:
            from numpy import sqrt
            host_symbol = self.ctx.wf_dict['host_symbol']
            dict_suggestion = self.inputs.distance_suggestion.get_dict()
            lattice = self.ctx.wf_dict['lattice']

            if lattice == 'fcc':
                suggestion_factor = sqrt(2)
            elif lattice == 'bcc':
                suggestion_factor = 2 / sqrt(3)
            else:
                return self.exit_codes.ERROR_NOT_SUPPORTED_LATTICE

            suggestion = dict_suggestion.get(lattice, dict_suggestion[host_symbol]).get(host_symbol, 4.0)

            self.ctx.wf_dict['latticeconstant'] = float(suggestion_factor * suggestion)

    def relax_needed(self):
        """
        Returns true if interlayer relaxation should be performed.
        """
        if self.ctx.eos_needed or 'eos_output' in self.inputs:
            if not self.ctx.eos_needed:
                eos_output = self.inputs.eos_output
            else:
                try:
                    eos_output = self.ctx.eos_wc.outputs.output_eos_wc_para
                except NotExistent:
                    return self.exit_codes.ERROR_EOS_FAILED

            self.ctx.scaling_param = eos_output.get_dict()['scaling_gs']

        return self.ctx.relax_needed

    def run_relax(self):
        """
        Optimize interlayer distance.
        """
        self.report('INFO: submit Relaxation WorkChain')
        inputs = {}
        inputs, error = self.prepare_relax()
        if error:
            return error
        res = self.submit(FleurBaseRelaxWorkChain, **inputs)
        self.to_context(relax_wc=res)

    def prepare_relax(self):
        """
        Initialise inputs for Relax workchain
        """
        inputs = AttributeDict(self.exposed_inputs(FleurBaseRelaxWorkChain, namespace='relax'))
        inputs.metadata.label = 'Relax_symmetric_film'
        inputs.metadata.description = 'The Relax workchain relaxing film structure'

        if self.ctx.eos_needed or 'eos_output' in self.inputs:
            if not self.ctx.eos_needed:
                eos_output = self.inputs.eos_output
            else:
                try:
                    eos_output = self.ctx.eos_wc.outputs.output_eos_wc_para
                except NotExistent:
                    return self.exit_codes.ERROR_EOS_FAILED
            # print(eos_output.get_dict())
            scaling_param = eos_output.get_dict()['scaling_gs']

            out_create_structure = create_film_to_relax(wf_dict_node=Dict(dict=self.ctx.wf_dict),
                                                        scaling_parameter=Float(scaling_param),
                                                        suggestion_node=self.inputs.distance_suggestion)
            inputs.scf.structure = out_create_structure['structure']
            substrate = out_create_structure['substrate']
            # TODO: error handling might be needed
            self.ctx.substrate = substrate.uuid  # can not store aiida data nodes directly in ctx.

            if not isinstance(inputs.scf.structure, StructureData):
                return inputs, inputs.scf.structure

        return inputs, None

    def make_magnetic(self):
        """
        Analuses outputs of previous steps and generated the final
        structure suitable for magnetic film calculations.
        """

        if 'optimized_structure' not in self.inputs:
            if not self.ctx.relax_wc.is_finished_ok:
                return self.exit_codes.ERROR_RELAX_FAILED

        if self.ctx.relax_needed:
            optimized_structure = self.ctx.relax_wc.outputs.optimized_structure
        else:
            optimized_structure = self.inputs.optimized_structure

        para_dict = {
            'total_number_layers': self.ctx.wf_dict['total_number_layers'],
            'num_relaxed_layers': self.ctx.wf_dict['num_relaxed_layers'],
            'z_coordinate_window': self.ctx.wf_dict['z_coordinate_window']
        }

        if not self.ctx.substrate:  # workchain was stated from remote->Relax or optimized_structure
            if 'optimized_structure' in self.inputs:
                self.ctx.substrate = find_substrate(structure=self.inputs['optimized_structure'])
            else:
                self.ctx.substrate = find_substrate(remote=self.inputs.relax.scf.remote_data)

        # to track the provenance from which structures it was created
        magnetic = magnetic_slab_from_relaxed_cf(optimized_structure, load_node(self.ctx.substrate),
                                                 Dict(dict=para_dict))
        magnetic.label = 'magnetic_structure'
        magnetic.description = ('Magnetic structure slab created within FleurCreateMagneticWorkChain, '
                                'created from : {} and {}'.format(optimized_structure.uuid, self.ctx.substrate))

        self.out('magnetic_structure', magnetic)


'''
@cf
def save_structure(structure):
    """
    Save a structure data node to provide correct provenance.
    """
    structure_return = structure.clone()
    return structure_return
'''


@cf
def magnetic_slab_from_relaxed_cf(optimized_structure, substrate, para_dict):
    """ calcfunction which wraps magnetic_slab_from_relaxed to keep provenance """
    from aiida_fleur.tools.StructureData_util import magnetic_slab_from_relaxed

    magnetic = magnetic_slab_from_relaxed(optimized_structure, substrate, **para_dict.get_dict())

    return magnetic


@cf
def create_substrate_bulk(wf_dict_node):
    """
    Calcfunction to create a bulk structure of a substrate.

    :params wf_dict: AiiDA dict node with at least keys lattice, host_symbol and latticeconstant
    (If they are not there, raises KeyError)
    Lattice key supports only fcc and bcc

    raises ExitCode 380, ERROR_NOT_SUPPORTED_LATTICE
    """

    from aiida.engine import ExitCode
    from ase.lattice.cubic import FaceCenteredCubic
    from ase.lattice.cubic import BodyCenteredCubic

    wf_dict = wf_dict_node.get_dict()
    lattice = wf_dict['lattice']
    if lattice == 'fcc':
        structure_factory = FaceCenteredCubic
    elif lattice == 'bcc':
        structure_factory = BodyCenteredCubic
    else:
        return ExitCode(380, 'ERROR_NOT_SUPPORTED_LATTICE', message='Specified substrate has to be bcc or fcc.')

    directions = [[1, 0, 0], [0, 1, 0], [0, 0, 1]]
    host_symbol = str(wf_dict['host_symbol'])
    latticeconstant = float(wf_dict['latticeconstant'])
    size = (1, 1, 1)
    structure = structure_factory(directions=directions,
                                  symbol=host_symbol,
                                  pbc=(1, 1, 1),
                                  latticeconstant=latticeconstant,
                                  size=size)

    return StructureData(ase=structure)


@cf
def create_film_to_relax(wf_dict_node, scaling_parameter, suggestion_node):
    """
    Create a film structure those interlayers will be relaxed.
    """
    from aiida_fleur.tools.StructureData_util import create_manual_slab_ase, center_film
    from aiida_fleur.tools.StructureData_util import adjust_film_relaxation, adjust_sym_film_relaxation
    from aiida_fleur.tools.StructureData_util import mark_fixed_atoms, has_z_reflection
    from aiida_fleur.tools.StructureData_util import define_AFM_structures

    # scaling_parameter = eos_output.get_dict()['scaling_gs']
    wf_dict = wf_dict_node.get_dict()
    scaling_parameter = float(scaling_parameter)

    miller = wf_dict['miller']
    directions = wf_dict['directions']
    host_symbol = wf_dict['host_symbol']
    latticeconstant = float(wf_dict['latticeconstant'] * scaling_parameter**(1 / 3.0))
    size = wf_dict['size']
    replacements = wf_dict['replacements']
    pop_last_layers = wf_dict['pop_last_layers']
    decimals = wf_dict['decimals']
    lattice = wf_dict['lattice']
    hold_layers = wf_dict['hold_layers']
    last_layer_factor = wf_dict['last_layer_factor']
    first_layer_factor = wf_dict['first_layer_factor']
    adjustment_needed = wf_dict['adjustment_needed']
    AFM_name = wf_dict['AFM_name']
    magnetic_layers = wf_dict['magnetic_layers']
    AFM_layer_positions = wf_dict['AFM_layer_positions']

    structure = create_manual_slab_ase(lattice=lattice,
                                       miller=miller,
                                       directions=directions,
                                       host_symbol=host_symbol,
                                       latticeconstant=latticeconstant,
                                       size=size,
                                       replacements=replacements,
                                       decimals=decimals,
                                       pop_last_layers=pop_last_layers)

    structure = StructureData(ase=structure)

    # define substrate in case of unknown surface, will be rewritten for known ones
    substrate = create_manual_slab_ase(lattice=lattice,
                                       miller=miller,
                                       directions=directions,
                                       host_symbol=host_symbol,
                                       latticeconstant=latticeconstant,
                                       size=(1, 1, 1),
                                       replacements=None,
                                       decimals=decimals)

    tmp_substrate = create_manual_slab_ase(lattice=lattice,
                                           miller=miller,
                                           directions=directions,
                                           host_symbol=host_symbol,
                                           latticeconstant=latticeconstant,
                                           size=(2, 2, 2),
                                           replacements=None,
                                           decimals=decimals)

    bond_length = find_min_distance_unary_structure(tmp_substrate)

    suggestion = suggestion_node.get_dict()

    if adjustment_needed:
        if has_z_reflection(structure):
            structure = adjust_sym_film_relaxation(structure, suggestion, host_symbol, bond_length, last_layer_factor)
            sym_film = True
        else:
            structure = adjust_film_relaxation(structure, suggestion, host_symbol, bond_length, last_layer_factor,
                                               first_layer_factor)
            sym_film = False

    try:
        if AFM_layer_positions:  # in this case use given layer positions, not adjusted ones (for AFM if FM is done)
            AFM_structure = AFM_layer_positions
        else:
            AFM_structure = structure
        structure, substrate = define_AFM_structures(AFM_structure,
                                                     lattice,
                                                     directions,
                                                     host_symbol,
                                                     replacements,
                                                     latticeconstant,
                                                     size,
                                                     decimals,
                                                     pop_last_layers,
                                                     AFM_name,
                                                     magnetic_layers,
                                                     sym_film=sym_film)
    except ValueError as err:  # do not mark spin-up spin-down atoms, something went wrong
        if AFM_name != 'FM':
            raise ValueError('Could not mark spin-up and spin-down atoms') from err

    structure = mark_fixed_atoms(structure, hold_layers)

    centered_structure = center_film(structure)

    return {'structure': centered_structure, 'substrate': StructureData(ase=substrate)}


def find_min_distance_unary_structure(tmp_substrate):
    """
    Finds a minimal distance beteween atoms in a unary structure.

    ..warning:

        Make sure tmp_substrate to have a non-primitive unit cell!

    """
    import numpy as np
    distance = tmp_substrate.get_all_distances()
    di = np.diag_indices(len(distance))
    distance[di] = 100
    bond_length = np.amin(distance)
    return bond_length


def find_substrate(remote=None, structure=None):
    """
    Finds the stored substrate structure.
    If remote is given, goes up and tri
    """
    from aiida_fleur.workflows.base_relax import find_inputs_relax
    from aiida.plugins import DataFactory

    FleurinpData = DataFactory('fleur.fleurinp')

    if remote:
        inputs = find_inputs_relax(remote)
    elif structure:
        from aiida.orm import WorkChainNode
        inc_nodes = structure.get_incoming().all()
        for link in inc_nodes:
            if isinstance(link.node, WorkChainNode):
                relax_wc = link.node
                break

        if 'scf__remote_data' in relax_wc.inputs:
            inputs = find_inputs_relax(relax_wc.inputs.scf__remote_data)
        else:
            return relax_wc.inputs.scf__structure.get_incoming().all()[0].node.get_outgoing().get_node_by_label(
                'substrate').uuid

    if isinstance(inputs, FleurinpData):
        raise ValueError('Did not expect to find Relax WC started from FleurinpData')
    else:
        orig_structure = inputs[0]
        return orig_structure.get_incoming().all()[0].node.get_outgoing().get_node_by_label('substrate').uuid



# -*- coding: utf-8 -*-
'mode': 'band',
'kpath': 'auto',  #seek (aiida), fleur (only Max4) or string to pass to ase
'klistname': 'path-3',
'kpoints_number': None,
'kpoints_distance': None,
'kpoints_explicit': None,  #dictionary containing a list of kpoints, weights
#and additional arguments to pass to set_kpointlist
'sigma': 0.005,
'emin': -0.50,
'emax': 0.90,
'add_comp_para': {
    'only_even_MPI': False,
    'max_queue_nodes': 20,
    'max_queue_wallclock_sec': 86400
},
'inpxml_changes': [],



# -*- coding: utf-8 -*-
{'atomtypes': [[
  {'atomic_number': 4, 'coreconfig': '(1s1/2)', 'element': 'Be', 'natoms': 1,
   'species': 'Be_corehole1', 'stateOccupation': [
      {'(1s1/2)': ['1.00000000', '.50000000']},
      {'(2p1/2)': ['.50000000', '.00000000']}], 'valenceconfig': '(2s1/2) (2p1/2)'},
  {'atomic_number': 4, 'coreconfig': '[He]', 'element': 'Be', 'natoms': 1,
   'species': 'Be-2', 'stateOccupation': [{'(2p1/2)': ['.00000000', '.00000000']}],
   'valenceconfig': '(2s1/2) (2p1/2)'},
  {'atomic_number': 4, 'coreconfig': '[He]', 'element': 'Be', 'natoms': 1,
   'species': 'Be-2', 'stateOccupation': [{'(2p1/2)': ['.00000000', '.00000000']}],
   'valenceconfig': '(2s1/2) (2p1/2)'},
  {'atomic_number': 4, 'coreconfig': '[He]', 'element': 'Be', 'natoms': 1,
   'species': 'Be-2', 'stateOccupation': [{'(2p1/2)': ['.00000000', '.00000000']}],
   'valenceconfig': '(2s1/2) (2p1/2)'}]], 'bandgap': [0.0004425914],
  'bandgap_units':'eV', 'binding_energy': [53.57027767044], 'corehole_type': 'valence',
  'binding_energy_units': 'eV', 'binding_energy_convention': 'negativ',
  'coreholes_calculated': 'Be1s', 'coreholes_calculated_details': '', 'coresetup': [],
  'errors': [], 'fermi_energy': [0.3138075709], 'fermi_energy_unit': 'eV',
  'reference_bandgaps': [0.0225936434], 'reference_coresetup': [],
  'successful': true, 'total_energy_all': [-1554.08485250996],
  'total_energy_all_units': 'eV', 'total_energy_ref': [-1607.6551301804],
  'total_energy_ref_units': 'eV', 'warnings': [], 'hints': [],
  'weighted_binding_energy': [107.14055534088], 'weighted_binding_energy_units': 'eV',
  'workflow_name': 'fleur_corehole_wc', 'workflow_version': '0.3.2'
}



###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/JuDFTteam/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################
"""
This module contains the parser for a inpgen calculation and methods for
parsing different files produced by inpgen.
"""

from aiida.parsers import Parser
from aiida.common.exceptions import NotExistent, InputValidationError, ValidationError

from aiida_fleur.data.fleurinp import FleurinpData
from aiida_fleur.calculation.fleurinputgen import FleurinputgenCalculation

import pprint
import re


class Fleur_inputgenParser(Parser):
    """
    This class is the implementation of the Parser class for the FLEUR inpgen.
    It takes the files received from an inpgen calculation and creates AiiDA
    nodes for the Database. From the inp.xml file a FleurinpData object is
    created, also some information from the out file is stored in a
    ParameterData node.
    """

    _setting_key = 'parser_options'

    def __init__(self, node):
        """
        Initialize the instance of Fleur_inputgenParser
        """
        super().__init__(node)

        # these files should be at least present after success of inpgen
        self._default_files = {FleurinputgenCalculation._OUTPUT_FILE_NAME, FleurinputgenCalculation._INPXML_FILE_NAME}
        self._other_files = {FleurinputgenCalculation._SHELLOUT_FILE_NAME}

    def parse(self, **kwargs):
        """
        Takes inp.xml generated by inpgen calculation and created an FleurinpData node.

        :return: a dictionary of AiiDA nodes to be stored in the database.
        """

        try:
            output_folder = self.retrieved
        except NotExistent:
            self.logger.error('No retrieved folder found')
            return self.exit_codes.ERROR_NO_RETRIEVED_FOLDER

        # check what is inside the folder
        list_of_files = output_folder.list_object_names()
        self.logger.info(f'file list {list_of_files}')

        errorfile = FleurinputgenCalculation._ERROR_FILE_NAME
        if errorfile in list_of_files:
            try:
                with output_folder.open(errorfile, 'r') as error_file:
                    error_file_lines = error_file.read()
            except OSError:
                self.logger.error(f'Failed to open error file: {errorfile}.')
                return self.exit_codes.ERROR_OPENING_OUTPUTS
            # if not empty, has_error equals True, prior fleur 32
            if error_file_lines:
                if isinstance(error_file_lines, bytes):
                    error_file_lines = error_file_lines.replace(b'\x00', b' ')
                else:
                    error_file_lines = error_file_lines.replace('\x00', ' ')
                if 'Run finished successfully' not in error_file_lines:
                    self.logger.warning(
                        f'The following was written into std error and piped to {errorfile}: \n {error_file_lines}')
                    self.logger.error('Inpgen calculation did not finish successfully.')

        #Check for a message if the given profile could not be found
        #This is only in the stdout at the moment
        shellout_file = FleurinputgenCalculation._SHELLOUT_FILE_NAME
        if FleurinputgenCalculation._SHELLOUT_FILE_NAME in list_of_files:
            try:
                with output_folder.open(shellout_file, 'r') as shellout:
                    shellout_file_lines = shellout.read()
            except OSError:
                self.logger.error(f'Failed to open error file: {shellout_file}.')
                return self.exit_codes.ERROR_OPENING_OUTPUTS
            if shellout_file_lines:
                if isinstance(shellout_file_lines, bytes):
                    shellout_file_lines = shellout_file_lines.replace(b'\x00', b' ')
                else:
                    shellout_file_lines = shellout_file_lines.replace('\x00', ' ')
                if 'Could not find profile' in shellout_file_lines:
                    m = re.search(r'Could not find profile ([^\s]+)', shellout_file_lines)
                    profile_name = 'NOT_FOUND'
                    if m is not None:
                        profile_name = m.group(1)[:-1]  #The output contains a dot directly after the name
                    self.logger.error('Inpgen code does not recognize the given profile %s', profile_name)
                    return self.exit_codes.ERROR_UNKNOWN_PROFILE.format(profile=profile_name)

        inpxml_file = FleurinputgenCalculation._INPXML_FILE_NAME
        if inpxml_file not in list_of_files:
            self.logger.error(f"XML inp not found '{inpxml_file}'")
            return self.exit_codes.ERROR_NO_INPXML

        for file1 in self._default_files:
            if file1 not in list_of_files:
                self.logger.error(f"Expected file '{file1}' not found in retrieved folder, it was probably "
                                  'not created by inpgen')
                return self.exit_codes.ERROR_MISSING_RETRIEVED_FILES

        try:
            fleurinp = FleurinpData(files=[])
            fleurinp.set_file(inpxml_file, node=output_folder)
        except InputValidationError as ex:
            self.logger.error(f'FleurinpData initialization failed: {str(ex)}')
            if fleurinp.parser_info == {}:
                self.logger.error('Parser output: No Output produced')
            else:
                self.logger.error(f'Parser output: {pprint.pformat(fleurinp.parser_info)}')
            return self.exit_codes.ERROR_FLEURINPDATA_INPUT_NOT_VALID
        except ValidationError as ex:
            self.logger.error(f'FleurinpData validation failed: {str(ex)}')
            return self.exit_codes.ERROR_FLEURINPDATA_NOT_VALID

        self.logger.info('FleurinpData initialized')
        self.out('fleurinpData', fleurinp)



###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/JuDFTteam/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################
'''
Module for the command line interface of AiiDA-FLEUR
'''
import click
import click_completion
import difflib

from aiida_fleur import __version__
from aiida.cmdline.params import options, types
from .launch import cmd_launch
from .data import cmd_data
from .workflows import cmd_workflow
from .visualization import cmd_plot
from .util import options as options_af

# Activate the completion of parameter types provided by the click_completion package
# for bash: eval "$(_AIIDA_FLEUR_COMPLETE=source aiida-fleur)"
click_completion.init()

# Instead of using entrypoints and directly injecting verdi commands into aiida-core
# we created our own separete CLI because verdi will prob change and become
# less material science specific


class MostSimilarCommandGroup(click.Group):
    """
    Overloads the get_command to display a list of possible command
    candidates if the command could not be found with an exact match.
    """

    def get_command(self, ctx, cmd_name):
        """
        Override the default click.Group get_command with one giving the user
        a selection of possible commands if the exact command name could not be found.
        """
        cmd = click.Group.get_command(self, ctx, cmd_name)

        # return the exact match
        if cmd is not None:
            return cmd

        matches = difflib.get_close_matches(cmd_name, self.list_commands(ctx), cutoff=0.5)

        if not matches:
            # single letters are sometimes not matched, try with a simple startswith
            matches = [c for c in sorted(self.list_commands(ctx)) if c.startswith(cmd_name)][:3]

        if matches:
            ctx.fail("'{cmd}' is not a aiida-fleur command.\n\n"
                     'The most similar commands are: \n'
                     '{matches}'.format(cmd=cmd_name, matches='\n'.join(f'\t{m}' for m in sorted(matches))))
        else:
            ctx.fail(f"'{cmd_name}' is not a aiida-fleur command.\n\nNo similar commands found.")

        return None


# Uncomment this for now, has problems with sphinx-click
#@click.command('aiida-fleur', cls=MostSimilarCommandGroup, context_settings={'help_option_names': ['-h', '--help']})
@click.group('aiida-fleur', context_settings={'help_option_names': ['-h', '--help']})
@options.PROFILE(type=types.ProfileParamType(load_profile=True))
# Note, __version__ should always be passed explicitly here,
# because click does not retrieve a dynamic version when installed in editable mode
@click.version_option(__version__, '-v', '--version', message='AiiDA-FLEUR version %(version)s')
def cmd_root(profile):  # pylint: disable=unused-argument
    """CLI for the `aiida-fleur` plugin."""


# To avoid circular imports all commands are not yet connected to the root
# but they have to be here because of bash completion on the other hand, this
# makes them not work with the difflib...
# see how aiida-core does it.

cmd_root.add_command(cmd_launch)
cmd_root.add_command(cmd_data)
cmd_root.add_command(cmd_workflow)
cmd_root.add_command(cmd_plot)



###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/JuDFTteam/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################
"""Command line utilities to create and inspect `StructureData` nodes."""
import click
from aiida.cmdline.params import options
from aiida.cmdline.utils import decorators, echo

from . import cmd_data


@click.group('structure')
def cmd_structure():
    """Commands to create and inspect `StructureData` nodes."""


cmd_data.add_command(cmd_structure)
# import filename
# import -N pk_fleurinp


@cmd_structure.command('import')
@click.argument('filename', type=click.Path(exists=True))
@click.option('--fleurinp/--no-fleurinp',
              default=False,
              show_default=True,
              help='Store also the fleurinp and the extractor calcfunction in the db.')
@options.DRY_RUN()
@decorators.with_dbenv()
def cmd_import(filename, dry_run, fleurinp):
    """
    Import a `StructureData` from a Fleur input file.

    FILENAME is the name/path of the inp.xml file to use.

    If you want to import a structure from any file type you can use
    'verdi data structure import -ase <filename>' instead.
    """
    from aiida_fleur.data.fleurinp import FleurinpData

    if not filename.endswith('.xml'):
        echo.echo_critical('Error: Currently, only StructureData from a inp.xml file can be extracted.')
    fleurinpd = FleurinpData(files=[filename])
    if not fleurinp or dry_run:
        structure = fleurinpd.get_structuredata_ncf()
    else:
        structure = fleurinpd.get_structuredata()
    formula = structure.get_formula()

    if dry_run:
        echo.echo_success(f'parsed structure with formula {formula}')
    else:
        structure.store()
        echo.echo_success(
            f'parsed and stored StructureData<{structure.pk}> with formula {formula}, also stored FleurinpData<{fleurinpd.pk}>'
        )



###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/JuDFTteam/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################
"""
This file contains a CalcJob that represents FLEUR calculation.
"""
import os
import io

from aiida.engine import CalcJob
from aiida.orm import Dict
from aiida.orm import RemoteData
from aiida.common.datastructures import CalcInfo, CodeInfo
from aiida.common.utils import classproperty
from aiida.common.exceptions import InputValidationError
from aiida.common.exceptions import UniquenessError
from aiida_fleur.data.fleurinp import FleurinpData
from aiida_fleur.calculation.fleurinputgen import FleurinputgenCalculation


class FleurCalculation(CalcJob):
    """
    A CalcJob class that represents FLEUR DFT calculation.
    For more information about the FLEUR-code family go to http://www.flapw.de/
    """

    ######### Only this should be to be maintained! #########

    # should a kpt node be used or fleur generate the mesh?
    _use_kpoints = False
    _INPXML_FILE_NAME = 'inp.xml'

    # Default input and output files
    _INPUT_FILE = 'inp.xml'
    _OUTPUT_FILE = 'out.xml'

    # these will be shown in AiiDA
    _OUTPUT_FILE_NAME = 'aiida.out'  # Shell output
    _INPUT_FILE_NAME = 'inp.xml'

    # needed for calc
    _OUTXML_FILE_NAME = 'out.xml'
    _INP_FILE_NAME = 'inp'
    _ENPARA_FILE_NAME = 'enpara'
    _SYMOUT_FILE_NAME = 'sym.out'
    _CDN1_FILE_NAME = 'cdn1'
    _SHELLOUTPUT_FILE_NAME = 'shell.out'
    _ERROR_FILE_NAME = 'out.error'
    # other
    _OUT_FILE_NAME = 'out'
    _CDNC_FILE_NAME = 'cdnc'  # core charge density
    _TIME_INFO_FILE_NAME = 'time.info'
    _KPTS_FILE_NAME = 'kpts'
    _QPTS_FILE_NAME = 'qpts'
    _PLOT_INP_FILE_NAME = 'plot_inp'
    _MIX_HISTORY_FILE_NAME = 'mixing_history*'
    _POT_FILE_NAME = 'pot*'
    _POT1_FILE_NAME = 'pottot'
    _POT2_FILE_NAME = 'potcoul'
    _STRUCTURE_FILE_NAME = 'struct.xcf'
    _STARS_FILE_NAME = 'stars'
    _WKF2_FILE_NAME = 'wkf2'
    _CDN_HDF5_FILE_NAME = 'cdn.hdf'
    _CDN_LAST_HDF5_FILE_NAME = 'cdn_last.hdf'

    # special out files
    _DOS_FILE_NAME = 'DOS.*'
    _DOS_MAX5_FILE_NAME = 'Local.*'
    _DOSINP_FILE_NAME = 'dosinp'
    _BAND_GNU_FILE_NAME = 'band.gnu'
    _BAND_FILE_NAME = 'bands.*'
    _BANDDOS_FILE_NAME = 'banddos.hdf'

    # helper files
    _FLEUR_WARN_ONLY_INFO_FILE_NAME = 'FLEUR_WARN_ONLY'
    _JUDFT_WARN_ONLY_INFO_FILE_NAME = 'JUDFT_WARN_ONLY'
    _QFIX_FILE_NAME = 'qfix'
    _USAGE_FILE_NAME = 'usage.json'

    # relax (geometry optimization) files
    _RELAX_FILE_NAME = 'relax.xml'

    # jij files
    _JENERG_FILE_NAME = 'jenerg'
    _MCINP_FILE_NAME = 'MCinp'
    _QPTSINFO_FILE_NAME = 'qptsinfo'
    _SHELL_FILE_NAME = 'shells'
    _JCONST_FILE_NAME = 'jconst'

    # files for lda+U
    _NMMPMAT_FILE_NAME = 'n_mmp_mat'
    _NMMPMAT_HDF5_FILE_NAME = 'n_mmp_mat_out'

    #files for crystal field calculations
    _CFDATA_HDF5_FILE_NAME = 'CFdata.hdf'

    #files for greensfunctions
    _GREENSF_HDF5_FILE_NAME = 'greensf.hdf'

    # files for hybrid functionals
    _COULOMB1_FILE_NAME = 'coulomb1'
    _MIXBAS_FILE_NAME = 'mixbas'
    _CMT_FIlE_NAME = 'cmt'
    _CZ_FILE_NAME = 'cz'
    _OLAP_FILE_NAME = 'olap'
    _VR0_FILE_NAME = 'vr0'

    # files non-collinear calculation
    _RHOMAT_INP_FILE_NAME = 'rhomat_inp'
    _RHOMAT_OUT_FILE_NAME = 'rhomat_out'
    _CDN_FILE_NAME = 'cdn'
    _DIROFMAG_FILE_NAME = 'dirofmag'

    # files for Wannier 90
    _W90KPTS_FILE_NAME = 'w90kpts'
    _PROJ_FILE_NAME = 'proj'
    _WANN_INP_FILE_NAME = 'wann_inp'
    _BKPTS_FILE_NAME = 'bkpts'
    _WFMMN_FILE_NAME = 'WF*.mmn'
    _WFAMN_FILE_NAME = 'WF*.amn'
    _WFWIN_FILE_NAME = 'WF*.win'
    _WFWOUT_FILE_NAME = 'WF*.wout'
    _UNK_FILE_NAME = 'UNK*'
    _KPTSMAP_FILE_NAME = 'kptsmap'
    _PROJGEN_INP_FILE_NAME = 'projgen_inp'
    _IONS_FILE_NAME = 'IONS'
    _POLARIZATION_OUT_FILE_NAME = 'polarization_out'
    _HOPPING_FILE_NAME = 'hopping.*'
    _WF1HSOMTX_FILE_NAME = 'WF1.hsomtx'
    _RSSOCMAT_FILE_NAME = 'rssocmat.1'
    _RSNABLA_FILE_NAME = 'rsnabla.*'
    _WFNABL_FILE_NAME = 'WF*.nabl'

    # copy file lists. I rather don not like this.
    # Might gives rise to a lot of possible errors, if files or not there,
    # or Fleur did not created same, or at some point they will not be
    # deleted remotely.

    # Policy
    # we store everything needed for a further run in the local repository
    # (inp.xml, cdn1), also all important results files.
    # these will ALWAYS be copied from the local repository to the maschine
    # If a parent calculation exists, other files will be copied remotely
    #######

    # all possible files first chargedensity
    _copy_filelist1 = [
        _INP_FILE_NAME, _ENPARA_FILE_NAME, _SYMOUT_FILE_NAME, _CDN1_FILE_NAME, _KPTS_FILE_NAME, _STARS_FILE_NAME,
        _WKF2_FILE_NAME
    ]

    # after inpgen, before first chargedensity
    _copy_filelist_inpgen = [_INPXML_FILE_NAME]

    # for after fleur SCF [name, destination_name]
    _copy_scf_noinp = [[_CDN1_FILE_NAME, _CDN1_FILE_NAME]]

    _copy_scf_noinp_hdf = [[_CDN_LAST_HDF5_FILE_NAME, _CDN_HDF5_FILE_NAME]]

    _copy_scf = [[_CDN1_FILE_NAME, _CDN1_FILE_NAME], [_INPXML_FILE_NAME, _INPXML_FILE_NAME]]

    _copy_scf_hdf = [[_CDN_LAST_HDF5_FILE_NAME, _CDN_HDF5_FILE_NAME], [_INPXML_FILE_NAME, _INPXML_FILE_NAME]]

    _copy_filelist_scf_remote = [_MIX_HISTORY_FILE_NAME]

    _copy_filelist3 = [
        _INP_FILE_NAME, _ENPARA_FILE_NAME, _SYMOUT_FILE_NAME, _CDN1_FILE_NAME, _KPTS_FILE_NAME, _STARS_FILE_NAME,
        _WKF2_FILE_NAME, _MIX_HISTORY_FILE_NAME, _OUT_FILE_NAME, _POT_FILE_NAME
    ]

    _copy_scf_ldau_nohdf = [[_CDN1_FILE_NAME, _CDN1_FILE_NAME], [_INPXML_FILE_NAME, _INPXML_FILE_NAME],
                            [_NMMPMAT_FILE_NAME, _NMMPMAT_FILE_NAME]]

    _copy_scf_ldau_noinp_nohdf = [[_CDN1_FILE_NAME, _CDN1_FILE_NAME], [_NMMPMAT_FILE_NAME, _NMMPMAT_FILE_NAME]]

    # files need for rerun
    _copy_filelist_dos = [_INPXML_FILE_NAME, _CDN1_FILE_NAME]

    _copy_filelist_band = [_INPXML_FILE_NAME, _POT_FILE_NAME, _CDN1_FILE_NAME]

    _copy_filelist_hybrid = []
    _copy_filelist_jij = []

    # possible settings_dict keys
    _settings_keys = [
        'additional_retrieve_list', 'remove_from_retrieve_list', 'additional_remotecopy_list',
        'remove_from_remotecopy_list', 'cmdline'
    ]

    @classmethod
    def define(cls, spec):
        super().define(spec)

        # spec.input('metadata.options.input_filename', valid_type=six.string_types,
        #            default=cls._INPXML_FILE_NAME)
        spec.input('metadata.options.output_filename', valid_type=str, default=cls._OUTXML_FILE_NAME)
        spec.input('metadata.options.use_kpoints', valid_type=bool, default=cls._use_kpoints)
        spec.input('metadata.options.parser_name', valid_type=str, default='fleur.fleurparser')

        # inputs
        spec.input('fleurinpdata',
                   valid_type=FleurinpData,
                   required=False,
                   help='Use a FleurinpData node that specifies the input parameters'
                   'usually copy from the parent calculation, basically makes'
                   'the inp.xml file visible in the db and makes sure it has '
                   'the files needed.')
        spec.input('parent_folder',
                   valid_type=RemoteData,
                   required=False,
                   help='Use a remote or local repository folder as parent folder '
                   '(also for restarts and similar). It should contain all the '
                   'needed files for a Fleur calc, only edited files should be '
                   'uploaded from the repository.')
        spec.input('settings',
                   valid_type=Dict,
                   required=False,
                   help='This parameter data node is used to specify for some '
                   'advanced features how the plugin behaves. You can add files'
                   'the retrieve list, or add command line switches, '
                   'for all available features here check the documentation.')

        # declare outputs of the calculation
        spec.output('output_parameters', valid_type=Dict, required=False)
        spec.output('output_params_complex', valid_type=Dict, required=False)
        spec.output('relax_parameters', valid_type=Dict, required=False)
        spec.output('error_params', valid_type=Dict, required=False)
        spec.default_output_node = 'output_parameters'

        # exit codes
        spec.exit_code(300, 'ERROR_NO_RETRIEVED_FOLDER', message='No retrieved folder found.')
        spec.exit_code(301, 'ERROR_OPENING_OUTPUTS', message='One of the output files can not be opened.')
        spec.exit_code(302, 'ERROR_FLEUR_CALC_FAILED', message='FLEUR calculation failed for unknown reason.')
        spec.exit_code(303, 'ERROR_NO_OUTXML', message='XML output file was not found.')
        spec.exit_code(304, 'ERROR_XMLOUT_PARSING_FAILED', message='Parsing of XML output file failed.')
        spec.exit_code(305, 'ERROR_RELAX_PARSING_FAILED', message='Parsing of relax XML output file failed.')
        spec.exit_code(310, 'ERROR_NOT_ENOUGH_MEMORY', message='FLEUR calculation failed due to lack of memory.')
        spec.exit_code(311,
                       'ERROR_VACUUM_SPILL_RELAX',
                       message='FLEUR calculation failed because an atom spilled to the'
                       'vacuum during relaxation')
        spec.exit_code(312, 'ERROR_MT_RADII', message='FLEUR calculation failed due to MT overlap.')
        spec.exit_code(313, 'ERROR_MT_RADII_RELAX', message='Overlapping MT-spheres during relaxation.')
        spec.exit_code(314, 'ERROR_DROP_CDN', message='Problem with cdn is suspected. Consider removing cdn')
        spec.exit_code(315,
                       'ERROR_INVALID_ELEMENTS_MMPMAT',
                       message='The LDA+U density matrix contains invalid elements.')
        spec.exit_code(316, 'ERROR_TIME_LIMIT', message='Calculation failed due to time limits.')
        spec.exit_code(318,
                       'ERROR_MISSING_DEPENDENCY',
                       message='Calculation failed due to missing dependency ({name}) for given calculation.')

    @classproperty
    def _get_output_folder(self):
        return './'

    def prepare_for_submission(self, folder):
        """
        This is the routine to be called when you make a FLEUR calculation.
        This routine checks the inputs and modifies copy lists accordingly.
        The standard files to be copied are given here.

        :param folder: a aiida.common.folders.Folder subclass where
                           the plugin should put all its files.
        """

        local_copy_list = []
        remote_copy_list = []
        remote_symlink_list = []
        mode_retrieved_filelist = []
        filelist_tocopy_remote = []
        settings_dict = {}

        has_fleurinp = False
        has_parent = False
        fleurinpgen = False
        copy_remotely = True
        with_hdf5 = False

        code = self.inputs.code

        codesdesc = code.description
        # TODO: ggf also check settings
        # In code description we write with what libs the code was compiled
        # we look for certain keywords in the description
        # also ggf, to be back comportable, the plugin should know the version number
        if codesdesc is not None:
            if 'hdf5' in codesdesc:
                with_hdf5 = True
            elif 'Hdf5' in codesdesc:
                with_hdf5 = True
            elif 'HDF5' in codesdesc:
                with_hdf5 = True
            else:
                with_hdf5 = False
        # a Fleur calc can be created from a fleurinpData alone
        # (then no parent is needed) all files are in the repo, but usually it is
        # a child of a inpgen calc or an other fleur calc (some or all files are
        # in a remote source). if the User has not changed something, the
        # calculation does not need theoretical a new FleurinpData it could use
        # the one from the parent, but the plug-in desgin is in a way that it has
        # to be there and it just copies files if changes occurred..

        if 'fleurinpdata' in self.inputs:
            fleurinp = self.inputs.fleurinpdata
        else:
            fleurinp = None

        if fleurinp is None:
            has_fleurinp = False
        else:
            has_fleurinp = True

        if 'parent_folder' in self.inputs:
            parent_calc_folder = self.inputs.parent_folder
        else:
            parent_calc_folder = None

        if parent_calc_folder is None:
            has_parent = False
            if not has_fleurinp:
                raise InputValidationError('No parent calculation found and no fleurinp data '
                                           'given, need either one or both for a '
                                           "'fleurcalculation'.")
        else:
            # extract parent calculation
            parent_calc = parent_calc_folder.creator
            has_parent = True

            # check that it is a valid parent
            # self._check_valid_parent(parent_calc)

            # if inpgen calc do
            # check if folder from db given, or get folder from rep.
            # Parent calc does not has to be on the same computer.

            # don't copy files, copy files locally
            copy_remotely = self.node.computer == parent_calc.computer
            if parent_calc.process_class not in (FleurCalculation, FleurinputgenCalculation):
                raise InputValidationError("parent_calc, must be either an 'inpgen calculation' or"
                                           f" a 'fleur calculation'. Got {parent_calc.process_class}")
            fleurinpgen = parent_calc.process_class is FleurinputgenCalculation

        # check existence of settings (optional)
        if 'settings' in self.inputs:
            settings_dict = self.inputs.settings.get_dict()
        else:
            settings_dict = {}

        # check for for allowed keys, ignore unknown keys but warn.
        for key in settings_dict.keys():
            if key not in self._settings_keys:
                self.logger.warning(
                    'settings dict key %s for Fleur calculation'
                    'not recognized, only %s are allowed.'
                    '', key, str(self._settings_keys))

        # TODO: Detailed check of FleurinpData
        # if certain files are there in fleurinpData
        # from where to copy

        # file copy stuff TODO check in fleur input
        if has_fleurinp:
            # add files belonging to fleurinp into local_copy_list
            for file in fleurinp.files:
                local_copy_list.append((fleurinp.uuid, file, file))
            modes = fleurinp.get_fleur_modes()

            # add files to mode_retrieved_filelist
            if with_hdf5 and (modes['band'] or modes['dos']):
                mode_retrieved_filelist.append(self._BANDDOS_FILE_NAME)

            if modes['band']:
                mode_retrieved_filelist.append(self._BAND_FILE_NAME)
                mode_retrieved_filelist.append(self._BAND_GNU_FILE_NAME)  #Should this be removed
            if modes['dos']:
                mode_retrieved_filelist.append(self._DOS_FILE_NAME)
                mode_retrieved_filelist.append(self._DOS_MAX5_FILE_NAME)
            if modes['relax']:
                # if l_f="T" retrieve relax.xml
                mode_retrieved_filelist.append(self._RELAX_FILE_NAME)
            if modes['ldau']:
                if with_hdf5:
                    mode_retrieved_filelist.append(self._NMMPMAT_HDF5_FILE_NAME)
                else:
                    mode_retrieved_filelist.append(self._NMMPMAT_FILE_NAME)
            if with_hdf5 and modes['greensf']:
                mode_retrieved_filelist.append(self._GREENSF_HDF5_FILE_NAME)
            if modes['cf_coeff']:
                if with_hdf5:
                    mode_retrieved_filelist.append(self._CFDATA_HDF5_FILE_NAME)
                else:
                    self.logger.warning('CF calculation without HDF5 not supported for automatic file retrieval.')
            if modes['force_theorem'] or modes['cf_coeff']:
                cdn_file = self._CDN_LAST_HDF5_FILE_NAME if with_hdf5 else self._CDN1_FILE_NAME
                settings_dict.setdefault('remove_from_retrieve_list', []).append(cdn_file)

            # if noco, ldau, gw...
            # TODO: check from where it was copied, and copy files of its parent
            # if needed

        if has_parent:
            # copy necessary files
            # TODO: check first if file exist and throw a warning if not
            outfolder_uuid = parent_calc.outputs.retrieved.uuid
            self.logger.info('out folder path %s', outfolder_uuid)

            outfolder_filenames = parent_calc.outputs.retrieved.list_object_names()
            has_nmmpmat_file = self._NMMPMAT_FILE_NAME in outfolder_filenames
            if (self._NMMPMAT_FILE_NAME in outfolder_filenames or \
                self._NMMPMAT_HDF5_FILE_NAME in outfolder_filenames):
                if has_fleurinp:
                    if self._NMMPMAT_FILE_NAME in fleurinp.files:
                        self.logger.warning('Ignoring {filename} from fleurinp. '
                                            'There is already an {filename} file '
                                            'for the parent calculation'.format(filename=self._NMMPMAT_FILE_NAME))
                        local_copy_list.remove((fleurinp.uuid, self._NMMPMAT_FILE_NAME, self._NMMPMAT_FILE_NAME))

            if fleurinpgen and (not has_fleurinp):
                for file1 in self._copy_filelist_inpgen:
                    if file1 not in outfolder_filenames:
                        raise InputValidationError(
                            f'File {file1} not found in parent folder but needed to start calculation')
                    local_copy_list.append((outfolder_uuid, file1, file1))
            elif not fleurinpgen and (not has_fleurinp):  # fleurCalc
                # need to copy inp.xml from the parent calc
                if with_hdf5:
                    copylist = self._copy_scf_hdf
                elif has_nmmpmat_file:
                    copylist = self._copy_scf_ldau_nohdf
                else:
                    copylist = self._copy_scf
                for file_orig, file_dest in copylist:
                    if file_orig not in outfolder_filenames:
                        if file_orig in (self._CDN1_FILE_NAME, self._CDN_LAST_HDF5_FILE_NAME):
                            raise InputValidationError(
                                f'File {file_orig} not found in parent folder but needed to start calculation'
                                'Make sure that the given Fleur code is correctly labelled with/without HDF5')
                        else:
                            raise InputValidationError(
                                f'File {file_orig} not found in parent folder but needed to start calculation')
                    local_copy_list.append((outfolder_uuid, file_orig, file_dest))
                # TODO: get inp.xml from parent fleurinpdata; otherwise it will be doubled in rep
            elif not fleurinpgen and has_fleurinp:
                # inp.xml will be copied from fleurinp
                if with_hdf5:
                    copylist = self._copy_scf_noinp_hdf
                elif has_nmmpmat_file:
                    copylist = self._copy_scf_ldau_noinp_nohdf
                else:
                    copylist = self._copy_scf_noinp
                for file_orig, file_dest in copylist:
                    if file_orig not in outfolder_filenames:
                        if file_orig in (self._CDN1_FILE_NAME, self._CDN_LAST_HDF5_FILE_NAME):
                            raise InputValidationError(
                                f'File {file_orig} not found in parent folder but needed to start calculation'
                                'Make sure that the given Fleur code is correctly labelled with/without HDF5')
                        else:
                            raise InputValidationError(
                                f'File {file_orig} not found in parent folder but needed to start calculation')
                    local_copy_list.append((outfolder_uuid, file_orig, file_dest))

            # TODO: not on same computer -> copy needed files from repository
            # if they are not there throw an error
            if copy_remotely:  # on same computer.
                filelist_tocopy_remote = filelist_tocopy_remote + self._copy_filelist_scf_remote
                # from settings, user specified
                # TODO: check if list?
                for file1 in settings_dict.get('additional_remotecopy_list', []):
                    filelist_tocopy_remote.append(file1)

                for file1 in settings_dict.get('remove_from_remotecopy_list', []):
                    if file1 in filelist_tocopy_remote:
                        filelist_tocopy_remote.remove(file1)

                for file1 in filelist_tocopy_remote:
                    remote_copy_list.append(
                        (parent_calc_folder.computer.uuid, os.path.join(parent_calc_folder.get_remote_path(),
                                                                        file1), self._get_output_folder))

                self.logger.info('remote copy file list %s', str(remote_copy_list))

        # create a JUDFT_WARN_ONLY file in the calculation folder
        with io.StringIO('/n') as handle:
            warn_only_filename = self._JUDFT_WARN_ONLY_INFO_FILE_NAME
            folder.create_file_from_filelike(handle, filename=warn_only_filename, mode='w')

        ########## MAKE CALCINFO ###########

        calcinfo = CalcInfo()

        calcinfo.uuid = self.uuid
        # Empty command line by default
        #cmdline_params = settings_dict.pop('CMDLINE', [])
        # calcinfo.cmdline_params = (list(cmdline_params)
        #                           + ["-in", self._INPUT_FILE_NAME])

        self.logger.info('local copy file list %s', str(local_copy_list))

        calcinfo.local_copy_list = local_copy_list
        calcinfo.remote_copy_list = remote_copy_list
        calcinfo.remote_symlink_list = remote_symlink_list

        # Retrieve by default the output file and the xml file
        retrieve_list = []
        retrieve_list.append(self._OUTXML_FILE_NAME)
        if has_fleurinp:
            for file in fleurinp.files:
                if file.endswith('.xml'):
                    retrieve_list.append(file)
        else:
            retrieve_list.append(self._INPXML_FILE_NAME)
        retrieve_list.append(self._SHELLOUTPUT_FILE_NAME)
        retrieve_list.append(self._ERROR_FILE_NAME)
        retrieve_list.append(self._USAGE_FILE_NAME)
        # retrieve_list.append(self._TIME_INFO_FILE_NAME)
        # retrieve_list.append(self._OUT_FILE_NAME)
        if with_hdf5:
            retrieve_list.append(self._CDN_LAST_HDF5_FILE_NAME)
        else:
            retrieve_list.append(self._CDN1_FILE_NAME)

        for mode_file in mode_retrieved_filelist:
            if mode_file not in retrieve_list:
                retrieve_list.append(mode_file)
        self.logger.info('retrieve_list: %s', str(retrieve_list))

        # user specific retrieve
        add_retrieve = settings_dict.get('additional_retrieve_list', [])
        self.logger.info('add_retrieve: %s', str(add_retrieve))
        for file1 in add_retrieve:
            retrieve_list.append(file1)

        remove_retrieve = settings_dict.get('remove_from_retrieve_list', [])
        for file1 in remove_retrieve:
            if file1 in retrieve_list:
                retrieve_list.remove(file1)

        calcinfo.retrieve_list = []
        for file1 in retrieve_list:
            calcinfo.retrieve_list.append(file1)

        codeinfo = CodeInfo()
        # should look like: codepath -xmlInput < inp.xml > shell.out 2>&1
        walltime_sec = self.node.get_attribute('max_wallclock_seconds')
        cmdline_params = []  # , "-wtime", "{}".format(walltime_sec)]"-xml"

        cmdline_params.append('-minimalOutput')

        if with_hdf5:
            cmdline_params.append('-last_extra')
            cmdline_params.append('-no_send')

        if walltime_sec:
            walltime_min = int(max(1, walltime_sec / 60))
            cmdline_params.append('-wtime')
            cmdline_params.append(f'{int(walltime_min)}')

        # user specific commandline_options
        for command in settings_dict.get('cmdline', []):
            cmdline_params.append(command)

        codeinfo.cmdline_params = list(cmdline_params)
        # + ["<", self._INPXML_FILE_NAME,
        # ">", self._SHELLOUTPUT_FILE_NAME, "2>&1"]
        codeinfo.code_uuid = code.uuid
        codeinfo.withmpi = self.node.get_attribute('withmpi')
        codeinfo.stdin_name = None  # self._INPUT_FILE_NAME
        codeinfo.stdout_name = self._SHELLOUTPUT_FILE_NAME
        #codeinfo.join_files = True
        codeinfo.stderr_name = self._ERROR_FILE_NAME

        calcinfo.codes_info = [codeinfo]

        return calcinfo



# -*- coding: utf-8 -*-
from aiida_fleur.workflows.banddos import FleurBandDosWorkChain
from aiida.orm import Dict, load_node
from aiida.engine import submit

fleur_code = load_node(FLEUR_PK)
inpgen_code = load_node(INPGEN_PK)
structure = load_node(STRUCTURE_PK)

wf_para = Dict(
    dict={
        'mode': 'band',
        'kpath': 'auto',  #seek (aiida), fleur (only Max4) or string to pass to ase
        'klistname': 'path-3',
        'kpoints_number': None,
        'kpoints_distance': None,
        'kpoints_explicit': None,  #dictionary containing a list of kpoints, weights
        #and additional arguments to pass to set_kpointlist
        'sigma': 0.005,
        'emin': -0.50,
        'emax': 0.90,
        'add_comp_para': {
            'only_even_MPI': False,
            'max_queue_nodes': 20,
            'max_queue_wallclock_sec': 86400
        },
        'inpxml_changes': [],
    })

wf_para_scf = Dict(
    dict={
        'fleur_runmax': 3,
        'density_converged': 0.001,
        'mode': 'density',
        'itmax_per_run': 30,
        'add_comp_para': {
            'only_even_MPI': False,
            'max_queue_nodes': 20,
            'max_queue_wallclock_sec': 86400
        }
    })

options = Dict(dict={
    'resources': {
        'num_machines': 1,
        'num_mpiprocs_per_machine': 2
    },
    'withmpi': True,
    'max_wallclock_seconds': 600
})

options_scf = Dict(dict={
    'resources': {
        'num_machines': 1,
        'num_mpiprocs_per_machine': 2
    },
    'withmpi': True,
    'max_wallclock_seconds': 600
})

calc_parameters = Dict(dict={'kpt': {'nkpts': 500, 'path': 'default'}})

inputs = {
    'scf': {
        'wf_parameters': wf_para_scf,
        'structure': structure,
        'calc_parameters': calc_parameters,
        'options': options_scf,
        'inpgen': inpgen_code,
        'fleur': fleur_code
    },
    'wf_parameters': wf_para,
    'fleur': fleur_code,
    'options': options
}

banddos_workchain = submit(FleurBandDosWorkChain, **inputs)



# -*- coding: utf-8 -*-
'sqa_ref': [0.7, 0.7],                  # sets theta and phi for the reference calc
'use_soc_ref': False,                   # True if reference calc should use SOC terms
'sqas_theta': [0.0, 1.57079, 1.57079],  # a list of theta values for the FT
'sqas_phi': [0.0, 0.0, 1.57079],        # a list of phi values for the FT
'add_comp_para': {
    'only_even_MPI': False,             # True if suppress parallelisation having odd number of MPI
    'max_queue_nodes': 20,              # Max number of nodes allowed (used by automatic error fix)
    'max_queue_wallclock_sec': 86400    # Max number of walltime allowed (used by automatic error fix)
    },
'soc_off': [],                          # a list of atom labels to switch off SOC term
'inpxml_changes': []                    # additional changes before the FT step



# -*- coding: utf-8 -*-
from aiida_fleur.workflows.ssdisp import FleurSSDispWorkChain
from aiida.orm import Dict, load_node

fleur_code = load_node(FLEUR_PK)
inpgen_code = load_node(INPGEN_PK)
structure = load_node(STRUCTURE_PK)

wf_para = Dict(dict={'points': 9,
                     'step': 0.002,
                     'guess': 1.00
                     })


wf_para_scf = Dict(dict={'fleur_runmax': 2,
                        'itmax_per_run': 120,
                        'density_converged': 0.2,
                        'mode': 'density'
               })


options_scf = Dict(dict={'resources': {'num_machines': 1, 'num_mpiprocs_per_machine': 8},
                         'queue_name': 'devel',
                         'custom_scheduler_commands': '',
                         'max_wallclock_seconds':  60*60})

inputs = {'scf': {
                  'wf_parameters': wf_para_scf,
                  'calc_parameters': parameters,
                  'options': options_scf,
                  'inpgen': inpgen_code,
                  'fleur': fleur_code
                 },
          'wf_parameters': wf_para,
          'structure': structure
}

SCF_workchain = submit(FleurSSDispWorkChain,
                       fleur=fleur_code,
                       inpgen=inpgen_code,
                       calc_parameters=calc_parameters,
                       structure=structure,
                       wf_parameters=wf_para,
                       options=options)



# -*- coding: utf-8 -*-
from aiida.orm import load_node, Dict
from aiida.engine import submit

from aiida_fleur.workflows.mae import FleurMaeWorkChain

structure = load_node(STRUCTURE_PK)
fleur_code = load_node(FLEUR_PK)
inpgen_code = load_node(INPGEN_PK)

wf_para = Dict(dict={'sqa_ref': [0.7, 0.7],
                     'use_soc_ref': False,
                     'sqas_theta': [0.0, 1.57079, 1.57079],
                     'sqas_phi': [0.0, 0.0, 1.57079],
                     'add_comp_para': {
                                        'only_even_MPI': False,
                                        'max_queue_nodes': 20,
                                        'max_queue_wallclock_sec': 86400
                                        },
                     'soc_off': [],
                     'inpxml_changes': [],
                     })

options = Dict(dict={'resources': {'num_machines': 1, 'num_mpiprocs_per_machine': 24},
                     'queue_name': 'devel',
                     'custom_scheduler_commands': '',
                     'max_wallclock_seconds':  60*60})


parameters = Dict(dict={'atom': {'element': 'Pt',
                                 'lmax': 8
                                 },
                        'atom2': {'element': 'Fe',
                                  'lmax': 8,
                                  },
                        'comp': {'kmax': 3.8,
                                 },
                        'kpt': {'div1': 20,
                                'div2': 24,
                                'div3': 1
                                }})

wf_para_scf = {'fleur_runmax': 2,
               'itmax_per_run': 120,
               'density_converged': 0.2,
               'mode': 'density'
               }

wf_para_scf = Dict(dict=wf_para_scf)

options_scf = Dict(dict={'resources': {'num_machines': 2, 'num_mpiprocs_per_machine': 24},
                         'queue_name': 'devel',
                         'custom_scheduler_commands': '',
                         'max_wallclock_seconds':  60*60})

inputs = {'scf': {'wf_parameters': wf_para_scf,
                  'structure': structure,
                  'calc_parameters': parameters,
                  'options': options_scf,
                  'inpgen': inpgen_code,
                  'fleur': fleur_code
                  },
          'wf_parameters': wf_para,
          'fleur': fleur_code,
          'options': options
          }


res = submit(FleurMaeWorkChain, **inputs)



###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/JuDFTteam/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################
'''
Here we collect physical constants which are used throughout the code
that way we ensure consitency
'''
# at some point one should import these from masci-tools, or better scipy or ase to ensure,
# that all judft plugins and tools take the same values, to make roundtrips consistent
# NIST https://physics.nist.gov/cgi-bin/cuu/Value?hrev
HTR_TO_EV = 27.211386245988  #(53)
BOHR_A = 0.5291772108
#Scipy bohr 5.29177210903e-11 m
#Scipy htr 27.211386245988 eV
# NIST BOHR 0.529177210903 #(80)
#https://physics.nist.gov/cgi-bin/cuu/Value?bohrrada0

#Fleur
#htr_eV   = 27.21138602
#bohr=0.5291772108
#bohrtocm=0.529177e-8
#pymatgen uses scipy.constants
#ase: Bohr 0.5291772105638411
#Hartree 27.211386024367243
#Rydberg 13.605693012183622
#1/Bohr
#1.8897261258369282
#aiida-core units:
#bohr_to_ang = 0.52917720859



###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/JuDFTteam/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################
"""
In this module is the plot_fleur method and its logic. The methods allows for the visualization on
every database node specifc to aiida-fleur. It depends on plot more general plot routines from
masci-tools which use matplotlib or bokeh as backend.
"""
# TODO but allow to optional parse information for saving and title,
#  (that user can put pks or structure formulas in there)
# INFO: AiiDAlab has implemented an extendable viewer class for data structures,
# which might be some point moved to aiida-core and extensible over entrypoints.

from pprint import pprint
import warnings
import numpy as np

from aiida.orm import load_node, Node, WorkChainNode, Dict
from aiida.common.exceptions import UniquenessError, NotExistent

from masci_tools.util.constants import HTR_TO_EV
from masci_tools.vis.common import set_defaults, set_default_backend, show_defaults, save_defaults, load_defaults, reset_defaults

__all__ = ('plot_fleur', 'set_defaults', 'set_default_backend', 'show_defaults', 'save_defaults', 'load_defaults',
           'reset_defaults')


def plot_fleur(*args, save=False, show_dict=False, show=True, backend=None, **kwargs):
    """
    Plot single or multiple Fleur WorkChainNodes. Can be started from the Workchain or output parameters (Dict) node.
    The following WorkChains are supported:

        - `FleurSCFWorkChain`: Plots the convergence of energy and distance for the calculation
        - `FleurEOSWorkChain`: Plots the total energy vs. volume for the calculated scalings
        - `FleurBandDosWorkChain`: Plot the bandstructure/DOS calculated
        - `FleurOrbControlWorkChain`: Plot the distribution of total energies for all run calculations

    This method takes any amount of AiiDA nodes as positional arguments and starts
    the standard visualisation either as single or multiple nodes in one visualisation (if supported and nodes provided in a list).
    i.e ``plot_fleur(123, [124,125], uuid)``

    :param save: bool if True the produced plots are saved to file
    :param show: bool if True the produced plots are immediately shown
    :param show_dict: bool if True the parameter/Dict node of the WorkChain is printed 9for single visualizations)
    :param backend: str, 'bokeh' or else 'matplotlib'/'mpl' (default None uses default in masci-tools routines)

    Kwargs are passed on to the plotting functions for each WorkChainNode

    :returns: a list of plot objects produced by the used plotting library for further modification or handling
              this might be used for a quick dashboard build.
    """

    all_plots = []
    for arg in args:
        if isinstance(arg, list):
            # try plot together
            p1 = plot_fleur_mn(arg, save=save, show=show, backend=backend, **kwargs)
            if len(p1) == 1:
                p1 = p1[0]
        else:
            #print(arg)
            # plot alone
            p1 = plot_fleur_sn(arg, show_dict=show_dict, show=show, save=save, backend=backend, **kwargs)
        all_plots.append(p1)

    return all_plots


def plot_fleur_sn(node, show_dict=False, **kwargs):
    """
    This methods takes any single AiiDA node and starts the standard visualization for
    if it finds one
    """

    plot_nodes, workflow_name, _ = classify_node(node)

    if show_dict:
        pprint(plot_nodes[0])

    try:
        plotf = FUNCTIONS_DICT[workflow_name]
    except KeyError as exc:
        raise ValueError('Sorry, I do not know how to visualize'
                         f' this node in plot_fleur: {workflow_name} {node}') from exc

    plot_result = plotf(*plot_nodes, **kwargs)

    return plot_result


def plot_fleur_mn(nodelist, **kwargs):
    """
    This methods takes any amount of AiiDA node as a list and starts
    the standard visualisation for it, if it finds one.

    Some nodes types it tries to display together if it knows how to.
    and if they are given as a list.

    param: save showed the plots be saved automatically

    """
    from collections import defaultdict
    from collections.abc import Iterable
    ###
    # Things to plot together
    all_nodes = defaultdict(list)
    node_labels = defaultdict(list)
    ###

    if not isinstance(nodelist, list):
        raise ValueError(f'The nodelist provided: {nodelist}, type {type(nodelist)} is not a list.')

    for node in nodelist:

        try:
            plot_nodes, workflow_name, label = classify_node(node)
        except ValueError as exc:
            warnings.warn(f'Failed to classify node {node}: {exc}\n'
                          'Skipping this one')
            continue

        all_nodes[workflow_name].append(plot_nodes)
        node_labels[workflow_name].append(label)

    all_plot_res = []
    for workflow_name, plot_nodes in all_nodes.items():
        try:
            plotf = FUNCTIONS_DICT[workflow_name]
        except KeyError:
            warnings.warn('Sorry, I do not know how to visualize'
                          f' these nodes in plot_fleur (multiplot): {workflow_name} {plot_nodes}')
            continue

        #Convert to tuple of lists
        plot_nodes = zip(*plot_nodes)
        plot_nodes = tuple(list(nodes) for nodes in plot_nodes)
        labels = node_labels[workflow_name]

        plot_res = plotf(*plot_nodes, labels=labels, **kwargs)
        if isinstance(plot_res, Iterable):
            all_plot_res.extend(plot_res)
        else:
            all_plot_res.append(plot_res)

    return all_plot_res


def classify_node(node):
    """
    Classify the given node and select, which nodes should be passed to the visualization
    function for the node.

    :param node: Aiida node or integer of the pk or str of the uuid

    :returns: tuple of nodes to pass to the corresponding visualization function,
              name of the workflow and label of the node
    """

    #Define any additional node hat should be passed to the plotting function
    ADDITIONAL_OUTPUTS = {
        'FleurBandDosWorkChain': ('last_calc_retrieved',),
        'FleurCFCoeffWorkChain': ('output_cfcoeff_wc_charge_densities', 'output_cfcoeff_wc_potentials')
    }

    if isinstance(node, (int, str)):
        node = load_node(node)

    if not isinstance(node, Node):
        raise ValueError(f'Given node {node} is not an AiiDA object')

    label = node.label

    parameter_node = None
    workchain_node = None
    if isinstance(node, WorkChainNode):
        workchain_node = node
        output_list = workchain_node.get_outgoing().all()
        for out_link in output_list:
            if 'output_' in out_link.link_label:
                if 'wc' in out_link.link_label or 'wf' in out_link.link_label:
                    if 'para' in out_link.link_label:  # We are just looking for parameter
                        #nodes, structures, bands, dos and so on we tread different
                        parameter_node = out_link.node  # we only visualize last output node
    elif isinstance(node, Dict):
        parameter_node = node

    if isinstance(parameter_node, Dict):
        parameter_dict = parameter_node.get_dict()
        workflow_name = parameter_dict.get('workflow_name', None)
    else:
        raise ValueError(f'I do not know how to visualize this node: {node}')

    add_outputs = ADDITIONAL_OUTPUTS.get(workflow_name, tuple())
    add_nodes = tuple()
    if add_outputs:
        if workchain_node is None:
            incoming = parameter_node.get_incoming(node_class=WorkChainNode).all()
            n_parents = len(incoming)
            if n_parents > 1:
                raise UniquenessError(f'Parameter node {parameter_node} has no unique WorkChainNode parent')
            if n_parents == 0:
                raise NotExistent(f'Parameter node {parameter_node} has no WorkChainNode parent')
            workchain_node = incoming[0].node

        add_nodes = tuple(workchain_node.get_outgoing().get_node_by_label(out_label) for out_label in add_outputs)

    outputs = (parameter_node,) + add_nodes

    return outputs, workflow_name, label


###########################
## general plot routine  ##
###########################


def plot_fleur_scf_wc(nodes, labels=None, save=False, show=True, backend='bokeh', **kwargs):
    """
    This methods takes an AiiDA output parameter node or a list from a scf workchain and
    plots number of iteration over distance and total energy
    """
    from masci_tools.vis.common import convergence_plot

    if not isinstance(nodes, list):
        nodes = [nodes]

    if labels is None:
        labels = [node.pk for node in nodes]

    all_energies = []
    all_distances = []
    all_iterations = []
    modes = []

    for node in nodes:
        iteration = []
        output_d = node.get_dict()
        total_energy = output_d.get('total_energy_all')
        if not total_energy:
            warnings.warn(f'No total energy data found, skip this node: {node}')
            continue

        distance = output_d.get('distance_charge_all')
        num_iterations = output_d.get('iterations_total')
        if not distance:
            warnings.warn(f'No distance_charge_all data found, skip this node: {node}')
            continue
        if not num_iterations:
            warnings.warn(f'No iteration_total data found, skip this node: {node}')
            continue

        mode = output_d.get('conv_mode')
        for i in range(1, len(total_energy) + 1):
            iteration.append(num_iterations - len(total_energy) + i)

        if len(distance) == 2 * len(total_energy):  # not sure if this is best solution
            # magnetic calculation, we plot only spin 1 for now.
            distance = [distance[j] for j in range(0, len(distance), 2)]

        all_energies.append(total_energy)
        all_iterations.append(iteration)
        all_distances.append(distance)
        modes.append(mode)

    add_args = {}
    if backend == 'bokeh':
        add_args['legend_label'] = labels
    else:
        add_args['plot_label'] = labels

    plot_res = convergence_plot(all_iterations,
                                all_distances,
                                all_energies,
                                show=show,
                                save_plots=save,
                                drop_last_iteration=any(mode == 'force' for mode in modes),
                                backend=backend,
                                **kwargs)
    return plot_res


def plot_fleur_dos_wc(node, labels=None, save=False, show=True, **kwargs):
    """
    This methods takes an AiiDA output parameter node from a density of states
    workchain and plots a simple density of states
    """
    from masci_tools.vis.plot_methods import plot_dos

    if labels is None:
        labels = []

    if isinstance(node, list):
        if len(node) > 2:
            return  # TODO
        else:
            node = node[0]

    output_d = node.get_dict()
    path_to_dosfile = output_d.get('dosfile', None)
    print(path_to_dosfile)
    if path_to_dosfile:
        data = np.loadtxt(path_to_dosfile, skiprows=0)

        energy = data[..., 0]
        dos_labels = ['Total', 'Interstitial', 'MT-Total']
        dos_data = [data[:, 1], data[:, 2], data[:, 1] - data[:, 2]]

        p1 = plot_dos(energy, dos_data, show=show, plot_labels=dos_labels)
    else:
        print('Could not retrieve dos file path from output node')

    return p1


def plot_fleur_eos_wc(nodes, labels=None, save=False, show=True, backend='bokeh', **kwargs):
    """
    This methods takes an AiiDA output parameter node from a equation of states
    workchain and plots a simple scaling vs volume plot
    """
    from masci_tools.vis.common import eos_plot

    if not isinstance(nodes, list):
        nodes = [nodes]

    energy = []
    scaling = []
    default_labels = []

    for i, nd in enumerate(nodes):
        outpara = nd.get_dict()
        volume_gs = outpara.get('volume_gs')
        scale_gs = outpara.get('scaling_gs')
        total_e = outpara.get('total_energy')
        if len(nodes) >= 2:
            total_e_norm = np.array(total_e) - total_e[0]
            energy.append(total_e_norm)
        else:
            energy.append(total_e)
        scaling.append(outpara.get('scaling'))
        default_labels.append(f'gs_vol: {volume_gs:.3} A^3, gs_scale {scale_gs:.3}, data {i}')

    labels = default_labels

    add_args = {}
    if backend == 'bokeh':
        add_args['legend_label'] = labels
    else:
        add_args['plot_label'] = labels

    plot_res = eos_plot(scaling, energy, show=show, save_plots=save, backend=backend, **add_args, **kwargs)

    return plot_res


def plot_fleur_band_wc(node, labels=None, save=False, show=True, **kwargs):
    """
    This methods takes an AiiDA output parameter node from a band structure
    workchain and plots a simple band structure
    """
    from masci_tools.vis.plot_methods import plot_bands

    if labels is None:
        labels = []

    if isinstance(node, list):
        if len(node) > 2:
            return  # TODO
        else:
            node = node[0]

    output_d = node.get_dict()
    path_to_bands_file = output_d.get('bandfile', None)
    print(path_to_bands_file)
    kpath = output_d.get('kpath', {})  #r"$\Gamma$": 0.00000, r"$H$" : 1.04590,
    #    r"$N$" : 1.78546, r"$P$": 2.30841, r"$\Gamma1$" : 3.21419, r"$N1$" : 3.95375} )
    if path_to_bands_file:
        data = np.loadtxt(path_to_bands_file, skiprows=0)
        kdata = data[..., 0]
        evdata = data[..., 1]
        p1 = plot_bands(kdata, evdata, special_kpoints=kpath)
    else:
        print('Could not retrieve dos file path from output node')

    return p1


def plot_fleur_banddos_wc(param_node,
                          file_node,
                          labels=None,
                          save=False,
                          show=True,
                          backend='bokeh',
                          hdf_recipe=None,
                          **kwargs):
    """
    This methods takes an AiiDA output parameter node and retrieved files from a banddos
    workchain and plots the bandstructure/DOS
    """
    from masci_tools.io.parsers.hdf5 import HDF5Reader
    from masci_tools.io.parsers.hdf5.recipes import FleurDOS, FleurBands
    from masci_tools.vis.fleur import plot_fleur_bands, plot_fleur_dos

    if isinstance(param_node, list):
        if len(param_node) > 2:
            return  # TODO
        else:
            param_node = param_node[0]
            file_node = file_node[0]

    output_d = param_node.get_dict()
    mode = output_d.get('mode')

    if mode is None:
        raise ValueError('Could not retrieve mode from output node')

    if hdf_recipe is None:
        if mode == 'dos':
            hdf_recipe = FleurDOS
        else:
            hdf_recipe = FleurBands

    if 'banddos.hdf' not in file_node.list_object_names():
        raise ValueError('No banddos.hdf file found')

    with file_node.open('banddos.hdf', mode='rb') as hdf_file:
        with HDF5Reader(hdf_file) as h5reader:
            data, attributes = h5reader.read(recipe=hdf_recipe)

    if mode == 'dos':
        plot_res = plot_fleur_dos(data, attributes, backend=backend, save_plots=save, show=show, **kwargs)
    else:
        plot_res = plot_fleur_bands(data, attributes, backend=backend, save_plots=save, show=show, **kwargs)

    return plot_res


def plot_fleur_relax_wc(node, labels=None, save=False, show=True, **kwargs):
    """
    This methods takes an AiiDA output parameter node from a relaxation
    workchain and plots some information about atom movements and forces
    """
    if labels is None:
        labels = []

    raise NotImplementedError


def plot_fleur_corehole_wc(nodes, labels=None, save=False, show=True, **kwargs):
    """
    This methods takes AiiDA output parameter nodes from a corehole
    workchain and plots some information about Binding energies
    """

    if labels is None:
        labels = []

    raise NotImplementedError


def plot_fleur_initial_cls_wc(nodes, labels=None, save=False, show=True, **kwargs):
    """
    This methods takes AiiDA output parameter nodes from a FleurInitialCLSWorkChain
    workchain and plots some information about corelevel shifts.
    (Spectra)
    """
    if labels is None:
        labels = []

    raise NotImplementedError


def plot_fleur_orbcontrol_wc(nodes,
                             labels=None,
                             save=False,
                             show=True,
                             line_labels=None,
                             backend='matplotlib',
                             **kwargs):
    """
    This methods takes AiiDA output parameter nodes from a orbcontrol
    workchain and plots the energy of the individual configurations.
    """
    from masci_tools.vis.common import scatter
    from itertools import chain

    if labels is None:
        labels = []

    if not isinstance(nodes, list):
        nodes = [nodes]

    offset = 0
    lines = []
    converged_configs = []
    converged_energy = []
    non_converged_configs = []
    non_converged_energy = []
    for node in nodes:
        outputs = node.get_dict()

        total_energy = outputs['total_energy']

        non_converged = outputs['non_converged_configs']
        converged = [i for i in outputs['successful_configs'] if i not in non_converged]

        converged_configs.extend(i + offset for i in converged)
        converged_energy.extend(total_energy[i] for i in converged)

        non_converged_configs.extend(i + offset for i in non_converged)
        non_converged_energy.extend(total_energy[i] for i in non_converged)

        offset += max(chain(converged, non_converged, outputs['failed_configs'])) + 1
        lines.append(offset - 0.5)

    #Convert to relative eV
    refE = min(converged_energy)
    converged_energy = np.array(converged_energy) - refE
    non_converged_energy = np.array(non_converged_energy) - refE

    converged_energy *= HTR_TO_EV
    non_converged_energy *= HTR_TO_EV

    if backend == 'matplotlib':
        kwargs.setdefault('plot_label', ['converged', 'not converged'])
    else:
        kwargs.setdefault('legend_label', ['converged', 'not converged'])

    kwargs.setdefault('xlabel', 'Configurations')
    kwargs.setdefault('ylabel', r'$E_{rel}$ [eV]')
    kwargs.setdefault('title', 'Results for orbcontrol node')
    kwargs.setdefault('legend_option', {'loc': 'upper right'})
    kwargs.setdefault('markersize', 10.0)
    kwargs.setdefault('legend', True)
    if len(lines) > 1:
        kwargs.setdefault('lines', {'vertical': lines[:-1]})

    p1 = scatter([converged_configs, non_converged_configs], [converged_energy, non_converged_energy],
                 color=['darkblue', 'darkred'],
                 save_plots=save,
                 show=show,
                 backend=backend,
                 **kwargs)

    if line_labels and backend == 'matplotlib':
        for label, pos in zip(line_labels, [0] + [p + 0.25 for p in lines]):
            p1.annotate(label, xy=(pos, 0.95), xycoords=('data', 'axes fraction'), ha='left', va='center', size=16)

    return p1


def plot_fleur_cfcoeff_wc(param_node,
                          cdn_node,
                          pot_node,
                          mode='calculation',
                          labels=None,
                          save=False,
                          show=True,
                          backend='matplotlib',
                          **kwargs):
    """
    Plot the CFCoeff workchain. Either plot the used potentials/charge densities or the angular dependence
    of the resulting potential
    """
    from masci_tools.tools.cf_calculation import plot_crystal_field_calculation, plot_crystal_field_potential
    from aiida_fleur.workflows.cfcoeff import reconstruct_cfcalculation, reconstruct_cfcoeffcients

    if isinstance(param_node, list):
        if len(param_node) > 2:
            return  # TODO
        else:
            param_node = param_node[0]
            cdn_node = cdn_node[0]
            pot_node = pot_node[0]

    output_d = param_node.get_dict()

    if mode not in ('calculation', 'potential'):
        raise ValueError(f'Invalid mode for plotting: {mode}')

    if backend != 'matplotlib':
        raise ValueError('Changing backend not yet implemented for CFCoeffWorkChain')

    plot_res = []
    if mode == 'potential':
        for atom_type in output_d['cf_coefficients_atomtypes']:
            coefficients = reconstruct_cfcoeffcients(output_d, atom_type)
            plot_res.append(plot_crystal_field_potential(coefficients, save=save, show=show, **kwargs))
    else:
        for atom_type in output_d['cf_coefficients_atomtypes']:
            cfcalc = reconstruct_cfcalculation(cdn_node, pot_node, atom_type)
            plot_res.append(plot_crystal_field_calculation(cfcalc, save=save, show=show, **kwargs))

    return plot_res


FUNCTIONS_DICT = {
    'fleur_scf_wc': plot_fleur_scf_wc,  #support of < 1.0 release
    'fleur_eos_wc': plot_fleur_eos_wc,  #support of < 1.0 release
    'FleurScfWorkChain': plot_fleur_scf_wc,
    'FleurEosWorkChain': plot_fleur_eos_wc,
    'fleur_dos_wc': plot_fleur_dos_wc,
    'fleur_band_wc': plot_fleur_band_wc,
    'FleurBandWorkChain': plot_fleur_band_wc,
    'FleurBandDosWorkChain': plot_fleur_banddos_wc,
    'FleurCFCoeffWorkChain': plot_fleur_cfcoeff_wc,
    #'fleur_corehole_wc' : plot_fleur_corehole_wc,  #support of < 1.5 release
    #'fleur_initial_cls_wc' : plot_fleur_initial_cls_wc,  #support of < 1.5 release
    #'FleurInitialCLSWorkChain' : plot_fleur_initial_cls_wc,
    #'FleurCoreholeWorkChain' :  plot_fleur_corehole_wc,
    'FleurOrbControlWorkChain': plot_fleur_orbcontrol_wc,
}



###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/JuDFTteam/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################
"""
Module with CLI commands to launch and inspect various aiida-fleur workchains.
"""
import click
from aiida.cmdline.params.types import ProcessParamType
from aiida.cmdline.params import arguments, options
from aiida.cmdline.utils import decorators, echo
from aiida import orm
from aiida_fleur.cmdline.util import options as options_fl


@click.group('workflow')
def cmd_workflow():
    """Commands to inspect aiida-fleur workchains."""


# general further commands for fleur workchains


@cmd_workflow.command('res')
@arguments.PROCESS('process', type=ProcessParamType()
                   )  #, type=WorkflowParamType(sub_classes=('aiida.node:process.workflow.workchain',)))
@click.option('--info/--no-info', default=False, help='Print an info header above each node.')
@click.option('-l', '--label', 'label', type=str, help='Print only output dicts with a certain link_label.')
@options_fl.SHOW()
@options.DICT_KEYS()
@options.DICT_FORMAT()
@decorators.with_dbenv()
def workchain_res(process, info, label, show, keys, fmt):
    """Print data from Dict nodes returned or created by any fleur process."""

    returned_dicts_info = []
    returned_dicts = []
    try:
        results = process.get_outgoing().all()
    except ValueError as exception:
        echo.echo_critical(str(exception))
    for result in results:
        if isinstance(result.node, orm.Dict):
            if label is not None:
                if label == result.link_label:
                    returned_dicts.append(result.node.get_dict())
                    returned_dicts_info.append(result)
            else:
                returned_dicts.append(result.node.get_dict())
                returned_dicts_info.append(result)

    for i, re_dict in enumerate(returned_dicts):
        if keys is not None:
            try:
                result = {k: re_dict[k] for k in keys}
            except KeyError as exc:
                echo.echo_critical(f"key '{exc.args[0]}' was not found in the results dictionary")
        else:
            result = re_dict
        if info:
            echo.echo(f'# Info: {returned_dicts_info[i].link_label} {returned_dicts_info[i].node} dict:')
        echo.echo_dictionary(result, fmt=fmt)


@cmd_workflow.command('inputdict')
@arguments.PROCESS('process', type=ProcessParamType()
                   )  #, type=WorkflowParamType(sub_classes=('aiida.node:process.workflow.workchain',)))
@click.option('--info/--no-info', default=False, help='Print an info header above each node.')
@click.option('-l', '--label', 'label', type=str, help='Print only output dicts with a certain link_label.')
@options_fl.SHOW()
@options.DICT_KEYS()
@options.DICT_FORMAT()
@decorators.with_dbenv()
def workchain_inputdict(process, info, label, show, keys, fmt):
    """Print data from Dict nodes input into any fleur process."""
    from aiida.orm import Dict

    returned_dicts_info = []
    returned_dicts = []
    try:
        results = process.get_incoming().all()
    except ValueError as exception:
        echo.echo_critical(str(exception))
    for result in results:
        if isinstance(result.node, orm.Dict):
            if label is not None:
                if label == result.link_label:
                    returned_dicts.append(result.node.get_dict())
                    returned_dicts_info.append(result)
            else:
                returned_dicts.append(result.node.get_dict())
                returned_dicts_info.append(result)

    for i, re_dict in enumerate(returned_dicts):
        if keys is not None:
            try:
                result = {k: re_dict[k] for k in keys}
            except KeyError as exc:
                echo.echo_critical(f"key '{exc.args[0]}' was not found in the results dictionary")
        else:
            result = re_dict
        if info:
            echo.echo(f'# Info: {returned_dicts_info[i].link_label} {returned_dicts_info[i].node} dict:')
        if show:
            echo.echo_dictionary(result, fmt=fmt)


'''
@cmd_workflow.command('gen_wf_para')
@arguments.ENTRYPOINTSTR('entrypoint', default='fleur.scf')
@options.SHOW()
@options.STORE()
@options.CHECK_EXISTENCE()
@options.KWARGS()
@decorators.with_dbenv()
def gen_wf_para_cmd(entrypoint, show, store, check_existence, kwargs):
    """
    Generates a default parameter wf parameter node for given entrypoint.
    """
    from aiida_fleur.tools.node_generators import generate_wf_para_node
    from aiida.plugins import entry_point
    try:
        wf = entry_point.load_entry_point('aiida.workflows', prefix)
    except ValueError:
        echo.echo('here1')
    try:
        wf = entry_point.load_entry_point('aiida.calculations', prefix)
    except ValueError:
        echo.echo('here1')

    wf_para = generate_wf_para_node(entrypoint=entrypoint, **kwargs)
    if store:
        wf_para.store()
        echo.echo('Created wf para node')
    else:
        echo.echo('Created wf para node')
    if show:
        echo.echo_dictionary(wf.para.get_dict())



@cmd_workflow.command('inputls')
def inputls_wc():
    """
    Prints verdi node show for all workchain inputs.
    """
    click.echo('verdi aiida-fleur scf res')

@cmd_workflow.command('show')
def show_wc():
    """
    Shows the node ans structure of a workchain.
    Similar to verdi node show
    """
    click.echo('verdi aiida-fleur scf show')


@cmd_workflow.command('list')
def list_wc():
    """
    Similar to the verdi process list command, but this has preset filters and
    displays also some specific information about fleur workchains
    """
    click.echo('verdi aiida-fleur scf list')
'''



###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/JuDFTteam/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################
"""
    In this module you find the workflow 'FleurMAEWorkChain' for the calculation of
    Magnetic Anisotropy Energy converging all the directions.
"""

import copy

from aiida.engine import WorkChain
from aiida.engine import calcfunction as cf
from aiida.orm import Dict
from aiida.common import AttributeDict

from aiida_fleur.workflows.scf import FleurScfWorkChain
from aiida_fleur.common.constants import HTR_TO_EV


class FleurMaeConvWorkChain(WorkChain):
    """
    This workflow calculates the Magnetic Anisotropy Energy of a structure.
    """

    _workflowversion = '0.2.1'

    _default_wf_para = {'sqas': {'label': [0.0, 0.0]}, 'soc_off': []}
    _default_options = {
        'resources': {
            'num_machines': 1,
            'num_mpiprocs_per_machine': 1
        },
        'max_wallclock_seconds': 6 * 60 * 60,
        'queue_name': '',
        'custom_scheduler_commands': '',
        'import_sys_environment': False,
        'environment_variables': {}
    }

    @classmethod
    def define(cls, spec):
        super().define(spec)
        spec.expose_inputs(FleurScfWorkChain, namespace='scf')
        spec.input('wf_parameters', valid_type=Dict, required=False)

        spec.outline(cls.start, cls.converge_scf, cls.get_results, cls.return_results)

        spec.output('out', valid_type=Dict)

        # exit codes
        spec.exit_code(230, 'ERROR_INVALID_INPUT_PARAM', message='Invalid workchain parameters.')
        spec.exit_code(343, 'ERROR_ALL_SQAS_FAILED', message='Convergence MAE calculation failed for all SQAs.')
        spec.exit_code(344, 'ERROR_SOME_SQAS_FAILED', message='Convergence MAE calculation failed for some SQAs.')

    def start(self):
        """
        Retrieve and initialize paramters of the WorkChain
        """
        self.report('INFO: started Magnetic Anisotropy Energy calculation'
                    ' convergence workflow version {}\n'.format(self._workflowversion))

        self.ctx.info = []
        self.ctx.warnings = []
        self.ctx.errors = []
        # defaults that will be written into the output node in case of failure
        # note: convergence branch always generates defaults inside get_results
        self.ctx.t_energydict = []
        self.ctx.mae_thetas = []
        self.ctx.mae_phis = []

        # initialize the dictionary using defaults if no wf paramters are given
        wf_default = copy.deepcopy(self._default_wf_para)
        if 'wf_parameters' in self.inputs:
            wf_dict = self.inputs.wf_parameters.get_dict()
        else:
            wf_dict = wf_default

        extra_keys = []
        for key in wf_dict.keys():
            if key not in wf_default.keys():
                extra_keys.append(key)
        if extra_keys:
            error = f'ERROR: input wf_parameters for MAE Conv contains extra keys: {extra_keys}'
            self.report(error)
            return self.exit_codes.ERROR_INVALID_INPUT_PARAM

        # extend wf parameters given by user using defaults
        for key, val in wf_default.items():
            wf_dict[key] = wf_dict.get(key, val)
        self.ctx.wf_dict = wf_dict

    def converge_scf(self):
        """
        Converge charge density with or without SOC.
        Depending on a branch of MAE calculation, submit a single Fleur calculation to obtain
        a reference for further force theorem calculations or
        submit a set of Fleur calculations to converge charge density for all given SQAs.
        """
        inputs = {}
        for key, soc in self.ctx.wf_dict['sqas'].items():
            inputs[key] = self.get_inputs_scf()
            inputs[key].calc_parameters['soc'] = {'theta': soc[0], 'phi': soc[1]}
            inputs[key].calc_parameters = Dict(dict=inputs[key].calc_parameters)
            res = self.submit(FleurScfWorkChain, **inputs[key])
            res.label = key
            self.to_context(**{key: res})

    def get_inputs_scf(self):
        """
        Initialize inputs for scf workflow
        """
        input_scf = AttributeDict(self.exposed_inputs(FleurScfWorkChain, namespace='scf'))

        if 'wf_parameters' not in input_scf:
            scf_wf_dict = {}
        else:
            scf_wf_dict = input_scf.wf_parameters.get_dict()

        if 'inpxml_changes' not in scf_wf_dict:
            scf_wf_dict['inpxml_changes'] = []

        # switch off SOC on an atom specie
        for atom_label in self.ctx.wf_dict['soc_off']:
            scf_wf_dict['inpxml_changes'].append(('set_species_label', {
                'at_label': atom_label,
                'attributedict': {
                    'special': {
                        'socscale': 0.0
                    }
                },
                'create': True
            }))

        input_scf.wf_parameters = Dict(dict=scf_wf_dict)

        if 'calc_parameters' in input_scf:
            calc_parameters = input_scf.calc_parameters.get_dict()
        else:
            calc_parameters = {}
        input_scf.calc_parameters = calc_parameters

        return input_scf

    def get_results(self):
        """
        Retrieve results of converge calculations
        """
        t_energydict = {}
        original_t_energydict = {}
        outnodedict = {}

        for label in self.ctx.wf_dict['sqas'].keys():
            calc = self.ctx[label]

            if not calc.is_finished_ok:
                message = f'One SCF workflow was not successful: {label}'
                self.ctx.warnings.append(message)
                continue

            try:
                outnodedict[label] = calc.outputs.output_scf_wc_para
            except KeyError:
                message = f'One SCF workflow failed, no scf output node: {label}. I skip this one.'
                self.ctx.errors.append(message)
                continue

            outpara = calc.outputs.output_scf_wc_para.get_dict()

            t_e = outpara.get('total_energy', 'failed')
            if not isinstance(t_e, float):
                message = f'Did not manage to extract float total energy from one SCF workflow: {label}'
                self.ctx.warnings.append(message)
                continue
            e_u = outpara.get('total_energy_units', 'Htr')
            if e_u in ['Htr', 'htr']:
                t_e = t_e * HTR_TO_EV
            t_energydict[label] = t_e

        if t_energydict:
            # Find a minimal value of MAE and count it as 0
            minenergy = min(t_energydict.values())

            for key, energy in t_energydict.items():
                original_t_energydict[key] = energy
                t_energydict[key] = energy - minenergy

        self.ctx.energydict = t_energydict
        self.ctx.original_energydict = original_t_energydict

    def return_results(self):
        """
        Retrieve results of converge calculations
        """

        failed_labels = []

        for label in self.ctx.wf_dict['sqas'].keys():
            if label not in self.ctx.energydict.keys():
                failed_labels.append(label)

        out = {
            'workflow_name': self.__class__.__name__,
            'workflow_version': self._workflowversion,
            # 'initial_structure': self.inputs.structure.uuid,
            'mae': self.ctx.energydict,
            'original_mae': self.ctx.original_energydict,
            'sqa': self.ctx.wf_dict['sqas'],
            'failed_labels': failed_labels,
            'mae_units': 'eV',
            'info': self.ctx.info,
            'warnings': self.ctx.warnings,
            'errors': self.ctx.errors
        }

        # create link to workchain node
        out = save_output_node(Dict(dict=out))
        self.out('out', out)

        if not self.ctx.energydict:
            return self.exit_codes.ERROR_ALL_SQAS_FAILED
        elif failed_labels:
            return self.exit_codes.ERROR_SOME_SQAS_FAILED

    def control_end_wc(self, errormsg):
        """
        Controlled way to shutdown the workchain. will initialize the output nodes
        The shutdown of the workchain will has to be done afterwards
        """
        self.report(errormsg)
        self.ctx.errors.append(errormsg)
        self.return_results()


@cf
def save_output_node(out):
    """
    This calcfunction saves the out dict in the db
    """
    out_wc = out.clone()
    return out_wc



###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/JuDFTteam/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################
"""
Module with CLI commands for various visualizations of data types.
"""
# if in the future there are futher sub commands of plot, the current plot
# command can become plot_fleur
import click
from aiida.cmdline.utils import decorators
from aiida.cmdline.params import arguments


@click.command('plot')
@arguments.NODES('nodes')
#                help='The pks for the nodes to be parsed to plot_fleur')
# type=click.Path(exists=True))
@click.option('-f', 'filename', type=click.File('r'), default=None)
@click.option('--save',
              type=click.BOOL,
              default=False,
              show_default=True,
              help='Should the result of plot_fleur be saved to a files.')
@click.option('--show/--no-show', default=True, show_default=True, help='Show the output of plot_fleur.')
@click.option('--show_dict/--no-show_dict', default=False, show_default=True, help='Show the output of plot_fleur.')
@click.option('--bokeh', 'backend', flag_value='bokeh')
@click.option('--matplotlib', 'backend', flag_value='matplotlib', default=True)
#@decorators.with_dbenv()
def cmd_plot(nodes, filename, save, show_dict, backend, show):
    """
    Invoke the plot_fleur command on given nodes
    """
    if backend != 'bokeh':
        # Try to set a working GUI backend for matplotlib
        # normally we assume to be on ipython which is not good.
        # The order is arbitrary.
        import matplotlib
        gui_env = [
            'WebAgg', 'GTK3Agg', 'GTK3Cairo', 'MacOSX', 'Qt4Agg', 'Qt4Cairo', 'Qt5Agg', 'Qt5Cairo', 'TkAgg', 'TkCairo',
            'WX', 'WXAgg', 'WXCairo', 'nbAgg', 'agg', 'cairo', 'pdf', 'pgf', 'ps', 'svg'
        ]
        for gui in gui_env:
            try:
                print('testing', gui)
                matplotlib.use(gui, force=True)
                from matplotlib import pyplot as plt
                break
            except ImportError as ex:
                print(ex)
                continue
        print('Using:', matplotlib.get_backend())

    from aiida_fleur.tools.plot.fleur import plot_fleur

    nodes = list(nodes)
    nodesf = []
    if filename is not None:
        nodesf = filename.read().split()
        filename.close()
    nodes = nodes + nodesf
    p = plot_fleur(nodes, save=save, show_dict=show_dict, backend=backend, show=show)



###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/JuDFTteam/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################
'''
This contains functions to generate certain node content, mostly dictionaries
with defaults.
These functions are usefull to expose one the commandline and to have defaults in one place
'''
# Todo: some of these can be moved to aiida-jutools
# since for kkr they might be the same
#todo, we allow kwargs for the nodes, maybe they should be validated
from aiida import orm
from aiida.plugins import WorkflowFactory
from aiida.orm import QueryBuilder


def generate_wf_option_node(computer=None, check_existence=True, **kwargs):
    """Create a option node for a certain workflow or calculation entry point.

    :param wf_entry_point: string the entry point to create the node for, default='fleur.scf'
    :param computer: dict {computername, queue} to provide computer dependend defaults
    :param kwargs: dict, further key word argument by which the node content will be updated

    :returns: AiiDA Dict node
    """
    option_node_dict = generate_wf_option_dict(computer=computer, **kwargs)
    option_node = orm.Dict(dict=option_node_dict)
    if check_existence:
        duplicate = QueryBuilder().append(orm.Dict, filters={'extras._aiida_hash': option_node._get_hash()}).first()  # pylint: disable=protected-access
    if duplicate:
        option_node = duplicate[0]

    return option_node


def generate_wf_option_dict(computer=None, protocol_file=None, **kwargs):
    """Create a option dict for a certain workflow or calculation entry point.

    :param computer: dict {computername, queue} to provide computer dependend defaults
    :param kwargs: dict, further key word argument by which the node content will be updated
    :param protocol_file: str, path to json file containing a set of default options

    :returns: python dict
    """
    # options usually do not differe between workflows, but depend on system size and machine.
    # Also per project, therefore it would be nice to allow to provide a file and with a set of defaults.
    # and this function should read them
    from aiida_fleur.common.defaults import default_options

    default_wf_dict = default_options.copy()
    #todo better rekursive merge?
    default_wf_dict.update(kwargs)

    return default_wf_dict


def generate_wf_para_dict(wf_entry_point='fleur.scf', **kwargs):
    """Create a wf parameter dict for a certain workflow or calculation entry point.

    :param wf_entry_point: string the entry point to create the node for, default='fleur.scf'
    :param kwargs: dict, further key word argument by which the node content will be updated

    :returns: python dict
    """

    wf = WorkflowFactory(wf_entry_point)
    default_wf_dict = wf._default_wf_para
    #todo better rekursive merge?
    default_wf_dict.update(kwargs)

    return default_wf_dict


def generate_wf_para_node(wf_entry_point='fleur.scf', check_existence=True, **kwargs):
    """Create a wf parameter node for a certain workflow or calculation entry point.

    :param wf_entry_point: string the entry point to create the node for, default='fleur.scf'
    :param kwargs: dict, further key word argument by which the node content will be updated

    :returns: AiiDA Dict node
    """
    wf_para_node_dict = generate_wf_para_dict(wf_entry_point=wf_entry_point, check_existence=check_existence, **kwargs)
    wf_para_node = orm.Dict(dict=wf_para_node_dict)
    if check_existence:
        duplicate = QueryBuilder().append(orm.Dict, filters={'extras._aiida_hash': wf_para_node._get_hash()}).first()  # pylint: disable=protected-access
    if duplicate:
        wf_para_node = duplicate[0]

    return wf_para_node



# pylint: disable=inconsistent-return-statements,protected-access,missing-docstring
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/JuDFTteam/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################
"""
In this module is the FleurinpModifier class, which is used to manipulate
FleurinpData objects in a way which keeps the provernance.
"""
import os
from lxml import etree
import warnings

from aiida.engine.processes.functions import calcfunction as cf
from aiida_fleur.data.fleurinp import FleurinpData
from aiida_fleur.tools.xml_aiida_modifiers import set_kpointsdata_f

from masci_tools.io.fleurxmlmodifier import ModifierTask, FleurXMLModifier


class FleurinpModifier(FleurXMLModifier):
    """
    A class which represents changes to the :class:`~aiida_fleur.data.fleurinp.FleurinpData` object.
    """

    _extra_functions = {'schema_dict': {'set_kpointsdata': set_kpointsdata_f}}

    def __init__(self, original):
        """
        Initiation of FleurinpModifier.
        """
        assert isinstance(original, FleurinpData), 'Wrong AiiDA data type'

        self._original = original
        self._other_nodes = {}

        super().__init__()

    @classmethod
    def apply_fleurinp_modifications(cls, new_fleurinp, modification_tasks):
        """
        Apply the modifications working directly on the cloned
        FleurinpData instance. The functions will warn the user if one of the methods
        executed here are after XML modifications in the task list, since this method will implictly
        reorder the order of the execution

        .. warning::
            These should be performed BEFORE the XML Modification methods
            in any of the functions doing modifcations (:py:meth:`FleurinpModifier.show()`,
            :py:meth:`FleurinpModifier.validate()`, :py:meth:`FleurinpModifier.freeze()`).

        :param new_fleurinp: The Fleurinpdata instance cloned from the original
        :param modification_tasks: a list of modification tuples
        """
        #These functions from the FleurinpData are supported
        fleurinp_mod_functions = {'set_file': new_fleurinp.set_file, 'del_file': new_fleurinp.del_file}

        warn = False
        for task in modification_tasks:
            if task.name in fleurinp_mod_functions:
                modification_tasks.remove(task)
                action = fleurinp_mod_functions[task.name]
                action(*task.args, **task.kwargs)
                if warn:
                    warnings.warn('The modification methods operating directly adding/removing files '
                                  'are performed before any XML modification methods')
            else:
                warn = True

    def get_avail_actions(self):
        """
        Returns the allowed functions from FleurinpModifier
        """
        outside_actions_fleurinp = {
            'set_kpointsdata': self.set_kpointsdata,
            'set_atomgr_att': self.set_atomgr_att,
            'set_atomgr_att_label': self.set_atomgr_att_label,
            'xml_set_attribv_occ': self.xml_set_attribv_occ,
            'xml_set_first_attribv': self.xml_set_first_attribv,
            'xml_set_all_attribv': self.xml_set_all_attribv,
            'xml_set_text_occ': self.xml_set_text_occ,
            'xml_set_text': self.xml_set_text,
            'xml_set_all_text': self.xml_set_all_text,
            'add_num_to_att': self.add_num_to_att,
            'set_file': self.set_file,
            'del_file': self.del_file
        }

        outside_actions_fleurxml = super().get_avail_actions().copy()

        return {**outside_actions_fleurxml, **outside_actions_fleurinp}

    def set_kpointsdata(self, kpointsdata_uuid, name=None, switch=False, kpoint_type='path'):
        """
        Appends a :py:func:`~aiida_fleur.tools.xml_aiida_modifiers.set_kpointsdata_f()` to
        the list of tasks that will be done on the FleurinpData.

        :param kpointsdata_uuid: node identifier or :class:`~aiida.orm.KpointsData` node to be written into ``inp.xml``
        :param name: str name to give the newly entered kpoint list (only MaX5 or later)
        :param switch: bool if True the entered kpoint list will be used directly (only Max5 or later)
        :param kpoint_type: str of the type of kpoint list given (mesh, path, etc.) only Max5 or later
        """
        from aiida.orm import KpointsData, load_node

        if isinstance(kpointsdata_uuid, KpointsData):
            kpointsdata_uuid = kpointsdata_uuid.uuid
        # Be more careful? Needs to be stored, otherwise we cannot load it

        num_nodes = sum('kpoints' in label for label in self._other_nodes) + 1
        node_label = f'kpoints_{num_nodes}'

        self._other_nodes[node_label] = load_node(kpointsdata_uuid)
        self._tasks.append(
            ModifierTask('set_kpointsdata',
                         args=(kpointsdata_uuid,),
                         kwargs={
                             'name': name,
                             'switch': switch,
                             'kpoint_type': kpoint_type
                         }))

    #Modification functions that were renamed in masci-tools

    def shift_value_species_label(self, *args, **kwargs):
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_names.shift_value_species_label()` to
        the list of tasks that will be done on the xmltree.

        :param atom_label: string, a label of the atom which specie will be changed. 'all' if set up all species
        :param attributename: name of the attribute to change
        :param value_given: value to add or to multiply by
        :param mode: 'rel' for multiplication or 'abs' for addition

        Kwargs if the attributename does not correspond to a unique path:
            :param contains: str, this string has to be in the final path
            :param not_contains: str, this string has to NOT be in the final path

        """
        if 'label' in kwargs:
            warnings.warn("The argument label is deprecated. Use 'atom_label' instead", DeprecationWarning)
            kwargs['atom_label'] = kwargs.pop('label')

        if 'att_name' in kwargs:
            warnings.warn("The argument att_name is deprecated. Use 'attributename' instead", DeprecationWarning)
            kwargs['attributename'] = kwargs.pop('att_name')

        if 'value' in kwargs:
            warnings.warn("The argument value is deprecated. Use 'value_given' instead", DeprecationWarning)
            kwargs['value_given'] = kwargs.pop('value')

        super().shift_value_species_label(*args, **kwargs)

    def set_species_label(self, *args, **kwargs):
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_names.set_species_label()` to
        the list of tasks that will be done on the xmltree.

        :param atom_label: string, a label of the atom which specie will be changed. 'all' to change all the species
        :param attributedict: a python dict specifying what you want to change.

        """
        if 'at_label' in kwargs:
            warnings.warn("The argument at_label is deprecated. Use 'atom_label' instead", DeprecationWarning)
            kwargs['atom_label'] = kwargs.pop('at_label')

        super().set_species_label(*args, **kwargs)

    def set_atomgr_att(self, *args, **kwargs):
        """
        Deprecated method for setting attributes on atomgroups
        """
        warnings.warn('This modification method is deprecated.'
                      "Use the 'set_atomgroup' method instead", DeprecationWarning)

        self.set_atomgroup(*args, **kwargs)

    def set_atomgr_att_label(self, *args, **kwargs):
        """
        Deprecated method for setting attributes on atomgroups identified by an atom label
        """
        warnings.warn('This modification method is deprecated.'
                      "Use the 'set_atomgroup_label' method instead", DeprecationWarning)

        if 'atom_label' not in kwargs:
            if len(args) == 2:
                kwargs['atom_label'], args = args[1], args[:1]
            elif len(args) > 2:
                kwargs['atom_label'], args = args[1], args[:1] + args[2:]

        self.set_atomgroup_label(*args, **kwargs)

    def xml_set_attribv_occ(self, *args, **kwargs):
        """
        Deprecated method for setting attributes for occurrences on a specific xpath
        """

        warnings.warn(
            'This modification method is deprecated.'
            "Use the 'xml_set_attrib_value_no_create' or"
            "'set_attrib_value' method instead", DeprecationWarning)

        if 'xpathn' in kwargs:
            kwargs['xpath'] = kwargs.pop('xpathn')

        occ = kwargs.pop('occ', None)
        kwargs.pop('create', None)

        self.xml_set_attrib_value_no_create(*args, **kwargs, occurrences=occ)

    def xml_set_first_attribv(self, *args, **kwargs):
        """
        Deprecated method for setting the first attribute on a specific xpath
        """

        warnings.warn(
            'This modification method is deprecated.'
            "Use the 'xml_set_first_attrib_value_no_create' with 'occurrences=0' or"
            "'set_first_attrib_value' method instead", DeprecationWarning)

        if 'xpathn' in kwargs:
            kwargs['xpath'] = kwargs.pop('xpathn')
        kwargs.pop('create', None)

        self.xml_set_attrib_value_no_create(*args, occurrences=0, **kwargs)

    def xml_set_all_attribv(self, *args, **kwargs):
        """
        Deprecated method for setting all attributes on a specific xpath
        """

        warnings.warn(
            'This modification method is deprecated.'
            "Use the 'xml_set_attrib_value_no_create' or"
            "'set_attrib_value' method instead", DeprecationWarning)

        if 'xpathn' in kwargs:
            kwargs['xpath'] = kwargs.pop('xpathn')
        kwargs.pop('create', None)

        self.xml_set_attrib_value_no_create(*args, **kwargs)

    def xml_set_text_occ(self, *args, **kwargs):
        """
        Deprecated method for setting texts for occurrences on a specific xpath
        """

        warnings.warn(
            'This modification method is deprecated.'
            "Use the 'xml_set_text_no_create' or"
            "'set_text' method instead", DeprecationWarning)

        if 'xpathn' in kwargs:
            kwargs['xpath'] = kwargs.pop('xpathn')
        occ = kwargs.pop('occ', None)
        kwargs.pop('create', None)
        kwargs.pop('place_index', None)
        kwargs.pop('tag_order', None)

        self.xml_set_text_no_create(*args, **kwargs, occurrences=occ)

    def xml_set_text(self, *args, **kwargs):
        """
        Deprecated method for setting attributes for occurrences on a specific xpath
        """

        warnings.warn(
            'This modification method is deprecated.'
            "Use the 'xml_set_text_no_create' with 'occurrences=0' or"
            "'set_first_text' method instead", DeprecationWarning)

        if 'xpathn' in kwargs:
            kwargs['xpath'] = kwargs.pop('xpathn')
        kwargs.pop('create', None)
        kwargs.pop('place_index', None)
        kwargs.pop('tag_order', None)

        self.xml_set_text_no_create(*args, occurrences=0, **kwargs)

    def xml_set_all_text(self, *args, **kwargs):
        """
        Deprecated method for setting attributes for occurrences on a specific xpath
        """

        warnings.warn(
            'This modification method is deprecated.'
            "Use the 'xml_set_text_no_create' or"
            "'set_text' method instead", DeprecationWarning)

        if 'xpathn' in kwargs:
            kwargs['xpath'] = kwargs.pop('xpathn')
        kwargs.pop('create', None)
        kwargs.pop('place_index', None)
        kwargs.pop('tag_order', None)

        self.xml_set_text_no_create(*args, **kwargs)

    def xml_create_tag(self, *args, **kwargs):
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_basic.xml_create_tag()` to
        the list of tasks that will be done on the xmltree.

        :param xpath: a path where to place a new tag
        :param element: a tag name or etree Element to be created
        :param place_index: defines the place where to put a created tag
        :param tag_order: defines a tag order
        :param occurrences: int or list of int. Which occurence of the parent nodes to create a tag.
                            By default all nodes are used.
        """
        self._validate_signature('xml_create_tag', *args, **kwargs)

        if 'element' in kwargs:
            element = kwargs
        else:
            element = args[1]

        if etree.iselement(element):
            warnings.warn('Creating a tag from a given etree Element is only supported via the show()'
                          'and validate() methods on the Fleurinpmodifier and cannot be used with freeze()')

        super().xml_create_tag(*args, **kwargs)

    def create_tag(self, *args, **kwargs):
        """
        Deprecation layer for create_tag if there are slashes in the first positional argument or xpath is is in kwargs.
        We know that it is the old usage.

        Appends a :py:func:`~masci_tools.util.xml.xml_setters_names.create_tag()` to
        the list of tasks that will be done on the xmltree.

        :param tag: str of the tag to create
        :param complex_xpath: an optional xpath to use instead of the simple xpath for the evaluation
        :param create_parents: bool optional (default False), if True and the given xpath has no results the
                               the parent tags are created recursively
        :param occurrences: int or list of int. Which occurence of the parent nodes to create a tag.
                            By default all nodes are used.

        Kwargs:
            :param contains: str, this string has to be in the final path
            :param not_contains: str, this string has to NOT be in the final path
        """

        old_interface = 'xpath' in kwargs
        if args:
            old_interface = old_interface or '/' in args[0]

        if old_interface:
            warnings.warn(
                "The 'create_tag' method no longer requires an explicit xpath. "
                'This Usage is deprecated. '
                "Use the 'xml_create_tag' method instead or only pass in the name of the tag, you want to use",
                DeprecationWarning)

            if 'xpath' in kwargs:
                xpath = kwargs.pop('xpath')
            else:
                xpath, args = args[0], args[1:]

            if 'newelement' in kwargs:
                element = kwargs.pop('newelement')
            else:
                element, args = args[0], args[1:]

            self.xml_create_tag(xpath, element, *args, **kwargs)
        else:
            tag = kwargs.get('tag')
            if tag is None:
                tag = args[0]

            if etree.iselement(tag):
                warnings.warn('Creating a tag from a given etree Element is only supported via the show()'
                              'and validate() methods on the Fleurinpmodifier and cannot be used with freeze()')

            super().create_tag(*args, **kwargs)

    def delete_tag(self, *args, **kwargs):
        """
        Deprecation layer for delete_tag if there are slashes in the first positional argument or xpath is is in kwargs.
        We know that it is the old usage.

        Appends a :py:func:`~masci_tools.util.xml.xml_setters_names.delete_tag()` to
        the list of tasks that will be done on the xmltree.

        :param tag: str of the tag to delete
        :param complex_xpath: an optional xpath to use instead of the simple xpath for the evaluation
        :param occurrences: int or list of int. Which occurence of the parent nodes to delete a tag.
                            By default all nodes are used.

        Kwargs:
            :param contains: str, this string has to be in the final path
            :param not_contains: str, this string has to NOT be in the final path
        """

        old_interface = 'xpath' in kwargs
        if args:
            old_interface = old_interface or '/' in args[0]

        if old_interface:
            warnings.warn(
                "The 'delete_tag' method no longer requires an explicit xpath. "
                'This Usage is deprecated. '
                "Use the 'xml_delete_tag' method instead or only pass in the name of the tag, you want to use",
                DeprecationWarning)

            if 'xpath' in kwargs:
                xpath = kwargs.pop('xpath')
            else:
                xpath, args = args[0], args[1:]

            self.xml_delete_tag(xpath, *args, **kwargs)
        else:
            super().delete_tag(*args, **kwargs)

    def delete_att(self, *args, **kwargs):
        """
        Deprecation layer for delete_att if there are slashes in the first positional argument or xpath is is in kwargs.
        We know that it is the old usage.

        Appends a :py:func:`~masci_tools.util.xml.xml_setters_names.delete_att()` to
        the list of tasks that will be done on the xmltree.

        :param tag: str of the attribute to delete
        :param complex_xpath: an optional xpath to use instead of the simple xpath for the evaluation
        :param occurrences: int or list of int. Which occurence of the parent nodes to delete a attribute.
                            By default all nodes are used.

        Kwargs:
            :param tag_name: str, name of the tag where the attribute should be parsed
            :param contains: str, this string has to be in the final path
            :param not_contains: str, this string has to NOT be in the final path
            :param exclude: list of str, here specific types of attributes can be excluded
                            valid values are: settable, settable_contains, other
        """

        old_interface = 'xpath' in kwargs
        if args:
            old_interface = old_interface or '/' in args[0]

        if old_interface:
            warnings.warn(
                "The 'delete_att' method no longer requires an explicit xpath. "
                'This Usage is deprecated. '
                "Use the 'xml_delete_att' method instead or only pass in the name of the attribute, you want to use",
                DeprecationWarning)

            if 'xpath' in kwargs:
                xpath = kwargs.pop('xpath')
            else:
                xpath, args = args[0], args[1:]

            self.xml_delete_att(xpath, *args, **kwargs)
        else:
            super().delete_att(*args, **kwargs)

    def xml_replace_tag(self, *args, **kwargs):
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_basic.xml_replace_tag()` to
        the list of tasks that will be done on the xmltree.

        :param xpath: a path to the tag to be replaced
        :param newelement: a new tag
        :param occurrences: int or list of int. Which occurence of the parent nodes to create a tag.
                            By default all nodes are used.
        """
        self._validate_signature('xml_replace_tag', *args, **kwargs)

        warnings.warn('Creating a tag from a given etree Element is only supported via the show()'
                      'and validate() methods on the Fleurinpmodifier and cannot be used with freeze()')

        super().xml_replace_tag(*args, **kwargs)

    def replace_tag(self, *args, **kwargs):
        """
        Deprecation layer for replace_tag if there are slashes in the first positional argument or xpath is is in kwargs.
        We know that it is the old usage.

        Appends a :py:func:`~masci_tools.util.xml.xml_setters_names.replace_tag()` to
        the list of tasks that will be done on the xmltree.

        :param tag: str of the tag to replace
        :param newelement: a new tag
        :param complex_xpath: an optional xpath to use instead of the simple xpath for the evaluation
        :param occurrences: int or list of int. Which occurence of the parent nodes to replace a tag.
                            By default all nodes are used.

        Kwargs:
            :param contains: str, this string has to be in the final path
            :param not_contains: str, this string has to NOT be in the final path
        """

        warnings.warn('Replacing a tag with a given etree Element is only supported via the show()'
                      'and validate() methods on the Fleurinpmodifier and cannot be used with freeze()')

        old_interface = 'xpath' in kwargs
        if args:
            old_interface = old_interface or '/' in args[0]

        if old_interface:
            warnings.warn(
                "The 'delete_att' method no longer requires an explicit xpath. "
                'This Usage is deprecated. '
                "Use the 'xml_delete_att' method instead or only pass in the name of the attribute, you want to use",
                DeprecationWarning)

            if 'xpath' in kwargs:
                xpath = kwargs.pop('xpath')
            else:
                xpath, args = args[0], args[1:]

            self.xml_replace_tag(xpath, *args, **kwargs)
        else:
            super().replace_tag(*args, **kwargs)

    def add_num_to_att(self, *args, **kwargs):
        """
        Deprecated method for adding a number to a attribute at a specific xpath
        """

        warnings.warn(
            'This modification method is deprecated.'
            "Use the 'add_number_to_attrib' or 'add_number_to_first_attrib' method instead", DeprecationWarning)

        #Since the new method takes only an attribute we extract the xpath and pass it in as complex_xpath

        if len(args) == 0:
            xpath = kwargs.pop('xpathn')
        else:
            xpath, args = args[0], args[1:]

        if 'occ' not in kwargs:
            self.add_number_to_first_attrib(*args, **kwargs, complex_xpath=xpath)
        else:
            occ = kwargs.pop('occ')
            self.add_number_to_attrib(*args, **kwargs, complex_xpath=xpath, occurrences=occ)

    def set_nmmpmat(self, *args, **kwargs):
        """
        Appends a :py:func:`~masci_tools.util.xml.xml_setters_nmmpmat.set_nmmpmat()` to
        the list of tasks that will be done on the xmltree.

        :param species_name: string, name of the species you want to change
        :param orbital: integer, orbital quantum number of the LDA+U procedure to be modified
        :param spin: integer, specifies which spin block should be modified
        :param state_occupations: list, sets the diagonal elements of the density matrix and everything
                          else to zero
        :param denmat: matrix, specify the density matrix explicitely
        :param phi: float, optional angle (radian), by which to rotate the density matrix before writing it
        :param theta: float, optional angle (radian), by which to rotate the density matrix before writing it
        """

        if 'occStates' in kwargs:
            warnings.warn("The argument occStates is deprecated. Use 'state_occupations' instead", DeprecationWarning)
            kwargs['state_occupations'] = kwargs.pop('occStates')

        super().set_nmmpmat(*args, **kwargs)

    def set_file(self, filename, dst_filename=None, node=None):
        """
        Appends a :py:func:`~aiida_fleur.data.fleurinp.FleurinpData.set_file()` to
        the list of tasks that will be done on the FleurinpData instance.

        :param filename: absolute path to the file or a filename of node is specified
        :param dst_filename: str of the filename, which should be used instead of the real filename
                             to store it
        :param node: a :class:`~aiida.orm.FolderData` node containing the file
        """
        from aiida.orm import load_node, Data

        node_uuid = None
        if node is not None:
            if isinstance(node, Data):
                node_uuid = node.uuid
            num_nodes = sum('folder' in label for label in self._other_nodes) + 1
            node_label = f'folder_{num_nodes}'
            # Be more careful? Needs to be stored, otherwise we cannot load it
            self._other_nodes[node_label] = load_node(node_uuid)

        self._tasks.append(
            ModifierTask('set_file', args=(filename,), kwargs={
                'dst_filename': dst_filename,
                'node': node_uuid
            }))

    def del_file(self, filename):
        """
        Appends a :py:func:`~aiida_fleur.data.fleurinp.FleurinpData.del_file()` to
        the list of tasks that will be done on the FleurinpData instance.

        :param filename: name of the file to be removed from FleurinpData instance
        """
        self._tasks.append(ModifierTask('del_file', args=(filename,), kwargs={}))

    def validate(self):
        """
        Extracts the schema-file.
        Makes a test if all the changes lead to an inp.xml file that is validated against the
        schema.

        :return: a lxml tree representing inp.xml with applied changes
        """

        new_fleurinp = self._original.clone()
        self.apply_fleurinp_modifications(new_fleurinp, self._tasks)

        xmltree, schema_dict = new_fleurinp.load_inpxml(remove_blank_text=True)
        develop_version = new_fleurinp.inp_version != schema_dict['inp_version']

        try:
            with new_fleurinp.open(path='n_mmp_mat', mode='r') as n_mmp_file:
                nmmplines = n_mmp_file.read().split('\n')
        except FileNotFoundError:
            nmmplines = None

        try:
            xmltree, nmmp = super().apply_modifications(xmltree, nmmplines, self._tasks)
        except etree.DocumentInvalid as exc:
            if not develop_version:
                raise
            else:
                new_fleurinp.logger.warning(f'Ignoring validation errors for modifications of develop version: \n{exc}')

        return xmltree

    def show(self, display=True, validate=False):
        """
        Applies the modifications and displays/prints the resulting ``inp.xml`` file.
        Does not generate a new
        :class:`~aiida_fleur.data.fleurinp.FleurinpData` object.

        :param display: a boolean that is True if resulting ``inp.xml`` has to be printed out
        :param validate: a boolean that is True if changes have to be validated

        :return: a lxml tree representing inp.xml with applied changes
        """

        if validate:
            xmltree = self.validate()
        else:
            new_fleurinp = self._original.clone()
            self.apply_fleurinp_modifications(new_fleurinp, self._tasks)

            xmltree, schema_dict = new_fleurinp.load_inpxml(remove_blank_text=True)
            try:
                with new_fleurinp.open(path='n_mmp_mat', mode='r') as n_mmp_file:
                    nmmplines = n_mmp_file.read().split('\n')
            except FileNotFoundError:
                nmmplines = None

            xmltree, nmmp = super().apply_modifications(xmltree, nmmplines, self._tasks, validate_changes=False)

        if display:
            xmltreestring = etree.tostring(xmltree, encoding='unicode', pretty_print=True)
            print(xmltreestring)
        return xmltree

    def freeze(self):
        """
        This method applies all the modifications to the input and
        returns a new stored fleurinpData object.

        :return: stored :class:`~aiida_fleur.data.fleurinp.FleurinpData` with applied changes
        """
        from aiida.orm import Dict
        modifications = Dict(dict={'tasks': self._tasks})
        modifications.description = 'Fleurinpmodifier Tasks and inputs of these.'
        modifications.label = 'Fleurinpdata modifications'
        # This runs in a inline calculation to keep provenance
        inputs = dict(original=self._original,
                      modifications=modifications,
                      metadata={
                          'label': 'fleurinp modifier',
                          'description': 'This calcfunction modified an Fleurinpdataobject'
                      },
                      **self._other_nodes)
        out = modify_fleurinpdata(**inputs)
        return out

    #Deactivate modify_xmlfile method from FleurXMLModifier (Only modify fleurinp)
    def modify_xmlfile(self, *args, **kwargs):  #pylint: disable=missing-function-docstring
        raise Exception(f'modify_xmlfile is disabled on {self.__class__.__name__}')


@cf
def modify_fleurinpdata(original, modifications, **kwargs):
    """
    A CalcFunction that performs the modification of the given FleurinpData and stores
    the result in a database.

    :param original: a FleurinpData to be modified
    :param modifications: a python dictionary of modifications in the form of {'task': ...}
    :param kwargs: dict of other aiida nodes to be linked to the modifications
    :returns new_fleurinp: a modified FleurinpData that is stored in a database
    """
    # copy
    # get schema
    # read in inp.xml
    # add modifications
    # validate
    # save inp.xml
    # store new fleurinp (copy)
    import tempfile
    from masci_tools.util.schema_dict_util import reverse_xinclude

    new_fleurinp = original.clone()

    modification_tasks = modifications.get_dict()['tasks']

    #We need to rebuild the namedtuples since the serialization for the calcufunction inputs
    #converts the namedtuples into lists
    modification_tasks = [ModifierTask(*task) for task in modification_tasks]

    FleurinpModifier.apply_fleurinp_modifications(new_fleurinp, modification_tasks)

    xmltree, schema_dict, included_tags = new_fleurinp.load_inpxml(remove_blank_text=True, return_included_tags=True)

    try:
        with new_fleurinp.open(path='n_mmp_mat', mode='r') as n_mmp_file:
            nmmplines = n_mmp_file.read().split('\n')
    except FileNotFoundError:
        nmmplines = None

    new_fleurtree, new_nmmplines = FleurinpModifier.apply_modifications(xmltree=xmltree,\
                                                                        nmmp_lines=nmmplines,\
                                                                        modification_tasks=modification_tasks,
                                                                        validate_changes=False)

    # To include object store storage this prob has to be done differently

    inpxmltree, includedtrees = reverse_xinclude(new_fleurtree, schema_dict, included_tags)

    new_fleurinp.del_file('inp.xml')
    with tempfile.TemporaryDirectory() as td:
        inpxml_path = os.path.join(td, 'inp.xml')
        inpxmltree.write(inpxml_path, encoding='utf-8', pretty_print=True)
        new_fleurinp.set_file(inpxml_path, 'inp.xml')

        for file_name, tree in includedtrees.items():
            file_path = os.path.join(td, file_name)
            tree.write(file_path, encoding='utf-8', pretty_print=True)
            new_fleurinp.set_file(file_path, file_name)

        if new_nmmplines is not None:
            n_mmp_path = os.path.join(td, 'n_mmp_mat')
            with open(n_mmp_path, 'w') as n_mmp_file:
                n_mmp_file.write('\n'.join(new_nmmplines))
            new_fleurinp.set_file(n_mmp_path, 'n_mmp_mat')

    # default label and description
    new_fleurinp.label = 'mod_fleurinp'
    new_fleurinp.description = 'Fleurinpdata with modifications (see inputs of modify_fleurinpdata)'

    return new_fleurinp
